# -*- coding: utf-8 -*-
#!/usr/bin/python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007      Thom Sturgill
# Copyright (C) 2007-2009 Brian G. Matherly
# Copyright (C) 2008-2011 Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2008      Jason Simanek
# Copyright (C) 2010      Jakim Friant
# Copyright (C) 2020-     Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Web Calendar generator.
"""

# ------------------------------------------------------------------------
# python modules
# ------------------------------------------------------------------------
import os
import datetime
import calendar  # Python module

# ------------------------------------------------------------------------
# Set up logging
# ------------------------------------------------------------------------
import logging

# ------------------------------------------------------------------------
# Gramps module
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.config import config
from gramps.gen.lib import Date, Name, NameType, Person
from gramps.gen.lib.date import Today
from gramps.gen.utils.alive import probably_alive
from gramps.gen.utils.db import get_death_or_fallback
from gramps.gen.utils.symbols import Symbols
from gramps.gen.datehandler import displayer as _dd

from gramps.gen.display.name import displayer as _nd

import gramps.plugins.lib.libholiday as libholiday
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import do_we_have_holidays, _WEB_EXT
from gramps.plugins.lib.libhtml import Html  # , xml_lang
from gramps.gui.pluginmanager import GuiPluginManager

from gramps.gen.lib.date import gregorian

# from gramps.plugins.lib.libhtmlbackend import HtmlBackend
# ------------------------------------------------------------------------
# constants
# ------------------------------------------------------------------------
_ = glocale.translation.sgettext
_LOG = logging.getLogger(".WebPage")

# full clear line for proper styling

FULLCLEAR = Html("div", class_="fullclear", inline=True)

# Calendar stylesheet names
_CALENDARSCREEN = "calendar-screen.css"
_CALENDARPRINT = "calendar-print.css"

PLUGMAN = GuiPluginManager.get_instance()
CSS = PLUGMAN.process_plugin_data("WEBSTUFF")

SCRIPT = """<script type="text/javascript">function currentmonth(y) {
var date = new Date();
var month = date.getMonth() + 1;
var url = y + "/" + month + "%s";
window.location.replace(url);
}
</script>
"""

CURRENT_DAY = """
<script type="text/javascript">
function CurrentDay() {
var date = new Date();
var months = "00" + (date.getMonth() + 1);
var days = "00" + date.getDate();
var day = days.substr(days.length-2);
var month = months.substr(months.length-2);
// Get the DOM reference
var elem = month+day;
var contentId = document.getElementById(elem);
contentId.style.background = "GreenYellow";
}
</script>
"""


# ------------------------------------------------------------------------
#
# WebCalReport
#
# ------------------------------------------------------------------------
class CalendarPage(BasePage):
    """
    Create WebCalReport object that produces the report.
    """

    def __init__(self, report, the_lang, the_title):
        """
        @param: report    -- The instance of the main report class
                             for this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)

        # class to do conversion of styled notes to html markup
        self.title_str = self._("Calendar")
        self.title_text = the_title
        self.name_format = report.options["name_format"]

        self.multiyear = False
        self.end_year = self.start_year = Today().get_year()

        self.maiden_name = report.options["maiden_name"]

        self.after_year = report.options["after_year"]
        self.alive = report.options["alive"]
        self.birthday = report.options["birthdays"]
        self.anniv = report.options["anniversaries"]
        self.death_anniv = report.options["death_anniv"]
        self.event_list = []
        self.current_month = None

        self._get_date = None
        self.fullyear = True
        self.makeoneday = report.options["makeoneday"]
        self.start_dow = report.options["start_dow"]

        self.warn_dir = True  # Only give warning once.

        self.link_to_narweb = True

        # self.calendar is a dict; key is the month number
        # Each entry in the dict is also a dict; key is the day number.
        # The day dict is a list of things to display for that day.
        # These things are: birthdays and anniversaries
        self.calendar = {}
        self.holidays = {}

        calendar.setfirstweekday(DOW_GRAMPS2ISO[self.start_dow])
        self.head = []

    def display_pages(self, the_lang, the_title):
        self.the_lang = the_lang
        self.the_title = the_title
        if the_lang:
            self.rlocale = self.report.set_locale(the_lang)
        else:
            self.rlocale = self.report.set_locale(self.report.options["trans"])
        self._ = self.rlocale.translation.sgettext
        self._get_date = self.rlocale.get_date
        self.title_text = the_title + " (" + self._("Calendar") + ")"
        self.title_str = self.title_text
        self.write_report()

    def add_day_item(
        self, text, year, month, day, event, age_at_death, dead_event_date
    ):
        """
        adds birthdays, anniversaries, and holidays to their perspective lists

        text -- line to be added
        year, month, day -- date to add the text to

        event -- one of 'BirthDay', 'Anniversary', 'Death' or 'Holiday'
        age_at_death -- The age in text. ie : 68 years, 6 months
        dead_event_date -- The date of the event used to calculate
                           the age_at_death
        """

        if year <= self.after_year:
            return

        # This may happen for certain "about" dates.
        # Use first day of the month
        if day == 0:
            day = 1

        # determine which dictionary to use???
        if event in ["Birthday", "Anniversary", "Death"]:
            month_dict = self.calendar.get(month, {})
        else:
            month_dict = self.holidays.get(month, {})
        day_list = month_dict.get(day, [])

        if month > 0:
            try:
                event_date = Date(year, month, day)
            except ValueError:
                event_date = Date.EMPTY
        else:
            event_date = Date.EMPTY  # Incomplete date....

        day_list.append((text, event, event_date, age_at_death, dead_event_date))
        month_dict[day] = day_list

        # determine which dictionary to add it to???
        if event in ["Birthday", "Anniversary", "Death"]:
            self.calendar[month] = month_dict
        else:
            self.holidays[month] = month_dict

    def __get_holidays(self, year):
        """Get the holidays for the specified country and year"""

        # _('translation')
        with self.r_user.progress(
            _("Web Calendar Report"),
            _("Calculating Holidays for year %04d") % year,
            365,
        ) as step:
            index = do_we_have_holidays(self.the_lang)
            if index is not None:
                holiday_table = libholiday.HolidayTable()
                country = holiday_table.get_countries()[index]
                holiday_table.load_holidays(year, country)
                for month in range(1, 13):
                    for day in range(1, 32):
                        holiday_names = holiday_table.get_holidays(month, day)
                        for holiday_name in holiday_names:
                            self.add_day_item(
                                holiday_name, year, month, day, "Holiday", None, None
                            )
                        step()

    def month_navigation(self, nr_up, year, currentsection):
        """
        Will create and display the navigation menu bar

        nr_up = number of directories up to reach root directory
        year = year being created
        currentsection = month name being created for proper CSS styling
        """
        navs = []

        navs.extend(
            (str(month), self.rlocale.date_displayer.short_months[int(month)], True)
            for month in range(1, 13)
        )

        # Add a link for year_glance() if requested
        navs.append(("fullyearlinked", self._("Full year at a Glance"), self.fullyear))

        # remove menu items if they are not True
        navs = [(u, n) for u, n, c in navs if c]

        # begin month subnavigation
        with Html("div", class_="wrapper", id="nav", role="navigation") as navigation:
            with Html("div", class_="container") as container:
                unordered = Html("ul", class_="menu " + self.dir)

                for url_fname, nav_text in navs:
                    # Note. We use '/' here because it is a URL, not a OS
                    # dependent pathname need to leave home link alone,
                    # so look for it ...
                    if nav_text != self._("NarrativeWeb Home"):
                        url_fname = url_fname.lower()
                    url = url_fname
                    add_subdirs = False
                    if not (url.startswith("http:") or url.startswith("/")):
                        add_subdirs = not any(url.endswith(ext) for ext in _WEB_EXT)

                    # whether to add subdirs or not???
                    if add_subdirs:
                        subdirs = [".."] * nr_up
                        subdirs.append("cal")
                        subdirs.append(str(year))
                        url = "/".join(subdirs + [url_fname])

                    if not _has_webpage_extension(url):
                        url += self.ext

                    # Figure out if we need <li class="CurrentSection"> or
                    # just plain <li>
                    check_cs = False
                    if url_fname == currentsection:
                        check_cs = 'class = "CurrentSection"'
                    elif url_fname == self.current_month:
                        check_cs = 'class = "CurrentSection"'

                    if url_fname == "fullyearlinked":
                        mytitle = self._("Full year at a Glance")
                    else:
                        mytitle = self._(url_fname)
                    hyper = Html("a", nav_text, href=url, name=url_fname, title=mytitle)

                    if check_cs:
                        unordered.extend(Html("li", hyper, attr=check_cs, inline=True))
                    else:
                        unordered.extend(Html("li", hyper, inline=True))
                container += unordered
            navigation += container
        return navigation

    def calendar_build(self, cal, year, month, clickable=False):
        """
        This does the work of building the calendar

        @param: cal - either "yg" year_glance(), or "wc" webcalendar()
        @param: year -- year being created
        @param: month - month number 1, 2, .., 12
        """

        date_displayer = self.rlocale.date_displayer
        # define names for long and short month names
        full_month_name = date_displayer.long_months[int(month)]
        abbr_month_name = date_displayer.short_months[int(month)]

        # dow (day-of-week) uses Gramps numbering, sunday => 1, etc
        start_dow = self.start_dow
        col2day = [(x - 1) % 7 + 1 for x in range(start_dow, start_dow + 7)]

        def get_class_for_daycol(col):
            """Translate a Gramps day number into a HTMLclass"""
            day = col2day[col]
            if day == 1:
                return "weekend sunday"
            elif day == 7:
                return "weekend saturday"
            return "weekday"

        def get_name_for_daycol(col):
            """Translate a Gramps day number into a HTMLclass"""
            day = col2day[col]
            return day_names[day]

        # Note. gen.datehandler has sunday => 1, monday => 2, etc
        # We slice out the first empty element.
        day_names = date_displayer.long_days  # use self._ldd.long_days when
        # set_locale is used ...

        def __get_previous_month_day(year, month, day_col):
            if month == 1:
                prevmonth = calendar.monthcalendar((year - 1), 12)
            else:
                prevmonth = calendar.monthcalendar(year, (month - 1))
            num_weeks = len(prevmonth)
            lastweek_prevmonth = prevmonth[(num_weeks - 1)]
            previous_month_day = lastweek_prevmonth[day_col]

            # return previous month day number based on day_col
            # day_col is based on range(0 - 6)
            return previous_month_day

        def __get_next_month_day(year, month, day_col):
            if month == 12:
                nextmonth = calendar.monthcalendar((year + 1), 1)
            else:
                nextmonth = calendar.monthcalendar(year, (month + 1))
            firstweek_nextmonth = nextmonth[0]
            next_month_day = firstweek_nextmonth[day_col]

            # return next month day number based on day_col
            # day_col is based on range(0 - 6)
            return next_month_day

        # Begin calendar head. We'll use the capitalized name, because here it
        # seems appropriate for most countries.
        month_name = full_month_name.capitalize()
        th_txt = month_name
        if cal == "wc":  # webcalendar()
            if not self.multiyear:
                th_txt = "%s %s" % (month_name, self._get_date(Date(year)))  # localized

        # begin calendar table and table head
        with Html(
            "table", class_="calendar", id=month_name, role="Calendar-Grid"
        ) as table:
            thead = Html("thead")
            table += thead

            if clickable:
                name = str(month) + self.ext
                linkable = Html("a", th_txt, href=name, name=name, title=th_txt)
                trow = Html("tr") + (
                    Html("th", linkable, class_="monthName", colspan=7, inline=True)
                )
                thead += trow
            else:
                if not self.multiyear:
                    self.end_year = self.start_year
                if month > 1:
                    full_month_name = str(month - 1)
                    url = full_month_name.lower() + self.ext
                    prevm = Date(int(year), int(month - 1), 0)
                    my_title = Html(
                        "a",
                        "\u276e",
                        href=url,
                        close=True,
                        title=date_displayer.display(prevm),
                    )
                elif self.multiyear and year > self.start_year:
                    full_month_name = str(12)
                    url = full_month_name.lower() + self.ext
                    dest = os.path.join("../", str(year - 1), url)
                    prevm = Date(int(year - 1), 12, 0)
                    my_title = Html(
                        "a",
                        "\u276e",
                        href=dest,
                        close=True,
                        title=date_displayer.display(prevm),
                    )
                else:
                    full_month_name = str(12)
                    url = full_month_name.lower() + self.ext
                    dest = os.path.join("../", str(self.end_year), url)
                    prevy = Date(self.end_year, 12, 0)
                    my_title = Html(
                        "a",
                        "\u276e",
                        href=dest,
                        close=True,
                        title=date_displayer.display(prevy),
                    )
                my_title += Html("</a>&nbsp;" + month_name + "&nbsp;")
                if month < 12:
                    full_month_name = str(month + 1)
                    url = full_month_name.lower() + self.ext
                    nextd = Date(int(year), int(month + 1), 0)
                    my_title += Html(
                        "a",
                        "\u276f",
                        href=url,
                        close=True,
                        title=date_displayer.display(nextd),
                    )
                elif self.multiyear and year < self.end_year:
                    full_month_name = str(1)
                    url = full_month_name.lower() + self.ext
                    dest = os.path.join("../", str(year + 1), url)
                    nextd = Date(int(year + 1), 1, 0)
                    my_title += Html(
                        "a",
                        "\u276f",
                        href=dest,
                        close=True,
                        title=date_displayer.display(nextd),
                    )
                else:
                    full_month_name = str(1)
                    url = full_month_name.lower() + self.ext
                    dest = os.path.join("../", str(self.start_year), url)
                    nexty = Date(self.start_year, 1, 0)
                    my_title += Html(
                        "a",
                        "\u276f",
                        href=dest,
                        close=True,
                        title=date_displayer.display(nexty),
                    )
                trow = Html("tr") + (
                    Html("th", my_title, class_="monthName", colspan=7, inline=True)
                )
                thead += trow
                trow = Html("tr") + (Html("th", "", colspan=7, inline=True))
                thead += trow

            # Calendar weekday names header
            trow = Html("tr")
            thead += trow

            for day_col in range(7):
                dayclass = get_class_for_daycol(day_col)
                dayname = self._(get_name_for_daycol(day_col))
                trow += Html("th", class_=dayclass, inline=True) + (
                    Html("abbr", dayname[0], title=dayname)
                )

            # begin table body
            tbody = Html("tbody")
            table += tbody

            # get first of the month and month information
            (dummy_current_date, current_ord, monthinfo) = get_first_day_of_month(
                year, month
            )

            # begin calendar table rows, starting week0
            nweeks = len(monthinfo)
            for week_row in range(0, nweeks):
                week = monthinfo[week_row]

                # if you look this up in wikipedia, the first week
                # is called week0
                trow = Html("tr", class_="week%02d" % week_row)
                tbody += trow

                # begin calendar day column
                for day_col in range(0, 7):
                    dayclass = get_class_for_daycol(day_col)

                    # day number, can also be a zero -- a day before or
                    # after month
                    day = week[day_col]

                    # start the beginning variable for <td>, table cell
                    tcell_id = "%02d%02d" % (month, day)

                    # add calendar date division
                    datediv = Html("div", day, class_="date", inline=True)
                    clickable = Html("div", datediv, class_="clickable", inline=True)

                    ### a day in the previous or next month ###
                    if day == 0:
                        # day in previous/ next month
                        specday = (
                            __get_previous_month_day(year, month, day_col)
                            if week_row == 0
                            else __get_next_month_day(year, month, day_col)
                        )

                        specclass = "previous " if week_row == 0 else "next "
                        specclass += dayclass

                        # continue table cell, <td>, without id tag
                        tcell = Html("td", class_=specclass, inline=True) + (
                            Html("div", specday, class_="date", inline=True)
                        )

                    # normal day number in current month
                    else:
                        thisday = datetime.date.fromordinal(current_ord)

                        # Something this month
                        if thisday.month == month:
                            holiday_list = self.holidays.get(month, {}).get(
                                thisday.day, []
                            )
                            bday_anniv_list = self.calendar.get(month, {}).get(
                                thisday.day, []
                            )

                            # date is an instance because of subtracting
                            # abilities in date.py
                            event_date = Date(thisday.year, thisday.month, thisday.day)

                            # get events for this day
                            day_list = get_day_list(
                                event_date,
                                holiday_list,
                                bday_anniv_list,
                                rlocale=self.rlocale,
                            )

                            # is there something this day?
                            if day_list:
                                hilightday = "highlight " + dayclass
                                tcell = Html("td", id=tcell_id, class_=hilightday)

                                # Year at a Glance
                                if cal == "yg":
                                    # make one day pages and hyperlink
                                    if self.makeoneday:
                                        # create yyyymmdd date string for
                                        # "One Day" calendar page filename
                                        fname_date = "%04d%02d%02d" % (year, month, day)
                                        fname_date += self.ext

                                        # create hyperlink to one_day()
                                        tcell += Html(
                                            "a", datediv, href=fname_date, inline=True
                                        )

                                        # only year_glance() needs this to
                                        # create the one_day() pages
                                        self.one_day(event_date, fname_date, day_list)

                                    # just year_glance(), but no one_day() pages
                                    else:
                                        # continue table cell, <td>,
                                        # without id tag
                                        tcell = Html(
                                            "td", class_=hilightday, inline=True
                                        ) + (
                                            # adds date division
                                            Html("div", day, class_="date", inline=True)
                                        )

                                # WebCal
                                else:
                                    # add date to table cell
                                    tcell += clickable

                                    # list the events
                                    unordered = Html("ul")
                                    clickable += unordered

                                    for (
                                        dummy_nyears,
                                        dummy_date,
                                        text,
                                        event,
                                        dummy_notused,
                                        dummy_notused,
                                    ) in day_list:
                                        unordered += Html(
                                            "li",
                                            text,
                                            inline=(
                                                False
                                                if (event == "Anniversary")
                                                else True
                                            ),
                                        )
                            # no events for this day
                            else:
                                # adds date division
                                date = Html("div", day, class_="date", inline=True)
                                # create empty day with date
                                tcell = Html("td", class_=dayclass, inline=True) + (
                                    # adds date division
                                    Html("div", date, class_="empty", inline=True)
                                )
                        # nothing for this month
                        else:
                            tcell = Html("td", class_=dayclass) + (
                                # adds date division
                                Html("div", day, class_="date", inline=True)
                            )

                    # attach table cell to table row
                    # close the day column
                    trow += tcell

                    # change day number
                    current_ord += 1

            if cal == "yg":
                for weeks in range(nweeks, 6):
                    # each calendar must have six weeks for proper styling
                    # and alignment
                    with Html("tr", class_="week%02d" % (weeks + 1)) as six_weeks:
                        tbody += six_weeks

                        for dummy_emptydays in range(7):
                            six_weeks += Html("td", class_="emptyDays", inline=True)

        # return calendar table to its callers
        return table

    def webcalendar(self, year):
        """
        This method provides information and header/ footer
        to the calendar month

        year -- year being created
        """

        # do some error correcting if needed
        if self.multiyear:
            if self.end_year < self.start_year:
                self.end_year = self.start_year

        # Number of directory levels up to get to self.html_dir / root
        if self.the_lang:
            nr_up = 2
        else:
            nr_up = 2

        with self.r_user.progress(
            _("Web Calendar Report"), _("Formatting months ..."), 12
        ) as step:
            for month in range(1, 13):
                self.current_month = str(month)
                cal_fname = "cal/%s/%s" % (str(year), str(month))
                open_file, sio = self.report.create_file(cal_fname)

                # Add xml, doctype, meta and stylesheets
                # body has already been added to webcal  already once
                result = self.write_header(self.title_text, cal=nr_up)
                webcal, head, dummy_body, outerwrapper = result

                # Create Month Navigation Menu
                # identify currentsection for proper highlighting
                currentsection = _dd.long_months[month]
                outerwrapper += self.month_navigation(nr_up, year, currentsection)

                # Set the current day background to GreenYellow
                # To do this, we force to load a bad image
                head += CURRENT_DAY
                current_day = Html("img", src="", onerror="CurrentDay()")
                outerwrapper += current_day

                # build the calendar
                content = Html("div", class_="content", id="WebCal")
                outerwrapper += content
                monthly_calendar = self.calendar_build("wc", year, month)
                content += monthly_calendar

                # create blank line for stylesheets
                # create footer division section
                footer = self.write_footer(None, cal=nr_up)
                outerwrapper += (FULLCLEAR, footer)

                # send calendar page to web output
                # and close the file
                self.xhtml_writer(webcal, open_file, sio, 0)

                step()

    def year_glance(self, year):
        """
        This method will create the Full Year At A Glance Page...
        year -- year being created
        """

        self.event_list = []
        prv = None
        nxt = None
        evdte = None
        for month in sorted(self.calendar):
            vals = sorted(self.calendar.get(month, {}))
            if month == 0:  # why ?
                continue
            for day in vals:
                event_date = "%04d%02d%02d" % (year, month, day)
                if evdte is None:
                    evdte = event_date
                elif nxt is None:
                    nxt = event_date
                    self.event_list.append((evdte, prv, nxt))
                else:
                    prv = evdte
                    evdte = nxt
                    nxt = event_date
                    self.event_list.append((evdte, prv, nxt))
        self.event_list.append((nxt, evdte, None))

        # Number of directory levels up to get to root
        if self.the_lang:
            nr_up = 2
        else:
            nr_up = 2

        # generate progress pass for "Year At A Glance"
        with self.r_user.progress(
            _("Web Calendar Report"), _("Creating Year At A Glance calendar"), 12
        ) as step:
            cal_fname = "cal/%s/fullyearlinked" % str(year)
            open_file, sio = self.report.create_file(cal_fname)

            # page title
            title = self._("%(year)d, At A Glance") % {"year": year}
            self.title_text = self.the_title + " (" + title + ")"
            self.title_str = self.title_text

            # Create page header
            # body has already been added to yearglance  already once
            result = self.write_header(cal_fname, cal=nr_up)
            yearglance, dummy_head, dummy_body, outerwrapper = result

            # Create Month Navigation Menu
            # identify currentsection for proper highlighting
            self.current_month = "0"
            outerwrapper += self.month_navigation(nr_up, year, "fullyearlinked")

            msg = self._(
                "This calendar is meant to give you access "
                "to all your data at a glance compressed into one "
                "page. Clicking on a date will take you to a page "
                "that shows all the events for that date, if there "
                "are any.\n"
            )

            # page description
            content = Html("div", class_="content", id="YearGlance")
            outerwrapper += content

            content += Html("p", msg, id="description")

            sav_cur_fname = self.report.cur_fname
            for month in range(1, 13):
                # build the calendar
                monthly_calendar = self.calendar_build(
                    "yg", year, month, clickable=True
                )
                content += monthly_calendar

                # increase progress bar
                step()

            # create blank line for stylesheets
            # write footer section
            footer = self.write_footer(None, cal=nr_up)
            outerwrapper += (FULLCLEAR, footer)

            # send calendar page to web output
            # and close the file
            self.report.cur_fname = sav_cur_fname
            self.xhtml_writer(yearglance, open_file, sio, 0)

    def one_day(self, event_date, fname_date, day_list):
        """
        This method creates the One Day page for "Year At A Glance"

        event_date -- date for the listed events

        fname_date -- filename date from calendar_build()

        day_list   -- a combination of both dictionaries to be able
                      to create one day
        nyears, date, text, event -- are necessary for figuring the age
        or years married for each year being created...
        """

        # Number of directory levels up to get to root
        if self.the_lang:
            nr_up = 2
        else:
            nr_up = 2

        # get year and month from event_date for use in this section
        year = event_date.get_year()
        month = event_date.get_month()

        cal_fname = "cal/%s/%s" % (str(year), fname_date)
        one_day_file, sio = self.report.create_file(cal_fname, ext="")

        # page title
        title = self._("One Day Within A Year")
        self.title_text = self.the_title + " (" + title + ")"
        self.title_str = self.title_text

        # Don't create the language menu for holidays.
        # This is to avoid 404 errors.
        for (
            dummy_nyears,
            dummy_date,
            text,
            event,
            dummy_age_at_death,
            dummy_dead_event_date,
        ) in day_list:
            self.not_holiday = False if event == "Holiday" else True

        # create page header
        result = self.write_header(cal_fname, cal=nr_up)
        oneday, dummy_head, dummy_body, outerwrapper = result

        # Create Month Navigation Menu
        # identify currentsection for proper highlighting
        currentsection = _dd.long_months[month]
        outerwrapper += self.month_navigation(nr_up, year, currentsection)

        # set date display as in user preferences
        content = Html("div", class_="content", id="OneDay")
        outerwrapper += content
        evt = fname_date[:8]
        found = (evt, None, None)
        for event in self.event_list:
            if event[0] == evt:
                found = event
                break
        my_title = Html()
        url = "#"
        if found[1] is not None:
            url = event[1] + self.ext
            prevd = Date(int(event[1][:4]), int(event[1][4:6]), int(event[1][6:]))
            my_title = Html("a", "\u276e", href=url, title=self.rlocale.get_date(prevd))
        else:
            my_title = Html("<em>&nbsp;&nbsp;</em>")
        my_title += Html("</a>")
        my_title += "&nbsp;&nbsp;"
        my_title += self.rlocale.date_displayer.display(event_date)
        my_title += "&nbsp;&nbsp;"
        if found[2] is not None:
            url = event[2] + self.ext
            nextd = Date(int(event[2][:4]), int(event[2][4:6]), int(event[2][6:]))
            my_title += Html(
                "a", "\u276f", href=url, title=self.rlocale.get_date(nextd)
            )
        else:
            my_title += Html("<b>&nbsp;&nbsp;</b>")
        content += Html("h3", my_title, inline=True)

        # list the events
        ordered = Html("ol")
        content += ordered
        for (
            dummy_nyears,
            dummy_date,
            text,
            event,
            dummy_age_at_death,
            dummy_dead_event_date,
        ) in day_list:
            ordered += Html(
                "li", text, inline=False if event == "Anniversary" else True
            )

        # create blank line for stylesheets
        # write footer section
        footer = self.write_footer(None, cal=nr_up)
        outerwrapper += (FULLCLEAR, footer)

        # send calendar page to web output
        # and close the file
        self.xhtml_writer(oneday, one_day_file, sio, 0)

    def get_name(self, person, maiden_name=None):
        """
        Return person's name, unless maiden_name given, unless married_name
        listed.

        person -- person to get short name from
        maiden_name  -- either a woman's maiden name or man's surname
        """
        # Get all of a person's names:
        primary_name = person.primary_name
        married_name = None
        names = [primary_name] + person.get_alternate_names()
        for name in names:
            if int(name.get_type()) == NameType.MARRIED:
                married_name = name
                break

        # Now, decide which to use:
        if maiden_name is not None:
            if married_name is not None:
                name = Name(married_name)
            else:
                name = Name(primary_name)
                surname_obj = name.get_primary_surname()
                surname_obj.set_surname(maiden_name)
        else:
            name = Name(primary_name)
        name.set_display_as(self.name_format)
        return _nd.display_name(name)

    def collect_data(self, this_year):
        """
        This method runs through the data, and collects the relevant dates
        and text.
        """
        db = self.r_db

        with self.r_user.progress(
            _("Web Calendar Report"),
            _("Reading database..."),
            len(self.report.obj_dict[Person]) + 1,
        ) as step:
            for person_handle in sorted(self.report.obj_dict[Person]):
                step()
                person = db.get_person_from_handle(person_handle)

                family_list = person.get_family_handle_list()
                birth_ref = person.get_birth_ref()
                birth_date = Date()
                if birth_ref:
                    birth_event = db.get_event_from_handle(birth_ref.ref)
                    birth_date = birth_event.get_date_object()

                death_ref = person.get_death_ref()
                person_death = Date()
                age_at_death = None
                if death_ref and birth_date:
                    death_event = db.get_event_from_handle(death_ref.ref)
                    death_date = death_event.get_date_object()
                    person_death = death_date
                    if birth_date != Date() and birth_date.is_valid() and death_date:
                        age_at_death = death_date - birth_date
                        age_at_death = age_at_death.format(dlocale=self.rlocale)

                # determine birthday information???
                if self.birthday and birth_date is not Date() and birth_date.is_valid():
                    birth_date = gregorian(birth_date)

                    year = birth_date.get_year()
                    month = birth_date.get_month()
                    day = birth_date.get_day()

                    # date to figure if someone is still alive
                    # current year of calendar, month nd day is their birth
                    # month and birth day
                    prob_alive_date = Date(this_year, month, day)
                    month, day = birth_date.anniversary(this_year)
                    # add some things to handle maiden name:
                    father_surname = None  # husband, actually
                    if person.gender == Person.FEMALE:
                        # get husband's last name:
                        if self.maiden_name in ["spouse_first", "spouse_last"]:
                            if family_list:
                                if self.maiden_name == "spouse_first":
                                    fhandle = family_list[0]
                                else:
                                    fhandle = family_list[-1]
                                fam = db.get_family_from_handle(fhandle)
                                father_handle = fam.get_father_handle()
                                mother_handle = fam.get_mother_handle()
                                if mother_handle == person.handle:
                                    if father_handle:
                                        father = db.get_person_from_handle(
                                            father_handle
                                        )
                                        if father is not None:
                                            father_surname = _regular_surname(
                                                person.gender, father.get_primary_name()
                                            )
                    short_name = self.get_name(person, father_surname)
                    alive = probably_alive(person, db, prob_alive_date)
                    if (self.alive and alive) or not self.alive:
                        # add link to NarrativeWeb
                        if self.link_to_narweb:
                            url = self.report.build_url_fname_html(
                                person.handle, "ppl", uplink=False
                            )
                            if self.usecms:
                                sub = ""
                            else:
                                sub = "/".join(([".."] * 2))
                            if sub:
                                url = url.replace("ppl", sub + "/ppl")
                            text = str(Html("a", short_name, href=url))
                        else:
                            text = short_name
                        if age_at_death is None:
                            self.add_day_item(
                                text,
                                year,
                                month,
                                day,
                                "Birthday",
                                age_at_death,
                                birth_date,
                            )
                        else:
                            self.add_day_item(
                                text,
                                year,
                                month,
                                day,
                                "Birthday",
                                age_at_death,
                                person_death,
                            )
                death_event = get_death_or_fallback(db, person)
                if death_event:
                    death_date = death_event.get_date_object()
                else:
                    death_date = None
                if self.death_anniv and death_date:
                    year = death_date.get_year() or this_year
                    month = death_date.get_month()
                    day = death_date.get_day()

                    short_name = self.get_name(person)
                    prob_alive_date = Date(this_year, month, day)
                    alive = probably_alive(person, db, prob_alive_date)
                    if (self.alive and alive) or not self.alive:
                        # add link to NarrativeWeb
                        if self.link_to_narweb:
                            hdle = person.handle
                            rbufh = self.report.build_url_fname_html
                            navpfx = rbufh(hdle, "ppl", uplink=False)
                            if self.usecms:
                                sub = ""
                            else:
                                sub = "/".join(([".."] * 2))
                            if sub:
                                navpfx = navpfx.replace("ppl", sub + "/ppl")
                            text = str(Html("a", short_name, href=navpfx))
                        else:
                            text = short_name
                        self.add_day_item(
                            text, year, month, day, "Death", age_at_death, death_date
                        )

                # add anniversary if requested
                if self.anniv:
                    for fhandle in family_list:
                        fam = db.get_family_from_handle(fhandle)
                        father_handle = fam.get_father_handle()
                        mother_handle = fam.get_mother_handle()
                        if father_handle == person.handle:
                            spouse_handle = mother_handle
                        else:
                            continue  # with next person if this was
                            # the marriage event
                        if spouse_handle:
                            spouse = db.get_person_from_handle(spouse_handle)
                            if spouse:
                                spouse_name = self.get_name(spouse)
                                short_name = self.get_name(person)
                                death_ref = spouse.get_death_ref()
                                spouse_death = Date()
                                if death_ref:
                                    death_event = db.get_event_from_handle(
                                        death_ref.ref
                                    )
                                    death_date = death_event.get_date_object()
                                    if death_date != Date() and death_date.is_valid():
                                        spouse_death = death_date
                            first_died = Date()
                            if person_death == Date():
                                first_died = spouse_death
                            elif spouse_death != Date():
                                first_died = (
                                    person_death
                                    if spouse_death > person_death
                                    else spouse_death
                                )
                            else:
                                first_died = person_death

                            # will return a marriage event or False if not
                            # married any longer
                            marriage_event = get_marriage_event(db, fam)
                            if marriage_event:
                                event_date = marriage_event.get_date_object()
                                if event_date is not Date() and event_date.is_valid():
                                    event_date = gregorian(event_date)
                                    year = event_date.get_year()
                                    month = event_date.get_month()
                                    day = event_date.get_day()

                                    month, day = event_date.anniversary(this_year)
                                    # date to figure if someone is still alive
                                    prob_alive_date = Date(this_year, month, day)
                                    wedding_age = None
                                    if first_died != Date():
                                        wedding_age = first_died - event_date
                                        wedding_age = wedding_age.format(
                                            dlocale=self.rlocale
                                        )
                                    divorce_event = get_divorce_event(db, fam)
                                    if divorce_event:
                                        d_date = divorce_event.get_date_object()
                                        if d_date is not Date() and d_date.is_valid():
                                            d_date = gregorian(d_date)
                                            if d_date != Date():
                                                w_age = d_date - event_date
                                                w_age = w_age.format(
                                                    dlocale=self.rlocale
                                                )
                                                wedding_age = w_age
                                                first_died = d_date

                                    if self.link_to_narweb:
                                        href = self.report.build_url_fname_html(
                                            spouse_handle, "ppl", uplink=False
                                        )
                                        if self.usecms:
                                            sub = ""
                                        else:
                                            sub = "/".join(([".."] * 2))
                                        if sub:
                                            href = href.replace("ppl", sub + "/ppl")
                                        spouse_name = str(
                                            Html("a", spouse_name, href=href)
                                        )
                                        href = self.report.build_url_fname_html(
                                            person.handle, "ppl", uplink=False
                                        )
                                        if sub:
                                            href = href.replace("ppl", sub + "/ppl")
                                        short_name = str(
                                            Html("a", short_name, href=href)
                                        )

                                    alive1 = probably_alive(person, db, prob_alive_date)
                                    alive2 = probably_alive(spouse, db, prob_alive_date)
                                    if first_died == Date():
                                        first_died = Date(0, 0, 0)
                                    if (
                                        self.alive and (alive1 or alive2)
                                    ) or not self.alive:
                                        spse = self._("%(spouse)s and" " %(person)s")
                                        text = spse % {
                                            "spouse": spouse_name,
                                            "person": short_name,
                                        }

                                        self.add_day_item(
                                            text,
                                            year,
                                            month,
                                            day,
                                            "Anniversary",
                                            wedding_age,
                                            first_died,
                                        )

    def write_report(self):
        """
        The short method that runs through each month and creates a page.
        """
        # get data from database for birthdays/ anniversaries
        self.collect_data(self.start_year)
        # initialize the holidays dict to fill:
        self.holidays = {}
        self.__get_holidays(self.start_year)

        if self.multiyear:
            # limit number of years to eighteen (18) years and only one row
            # of years
            nyears = (self.end_year - self.start_year) + 1
            num_years = nyears if 0 < nyears < 19 else 18

            for cal_year in range(self.start_year, (self.start_year + num_years)):
                # create webcalendar() calendar pages
                self.webcalendar(cal_year)

                # create "Year At A Glance" and
                # "One Day" calendar pages
                if self.fullyear:
                    self.year_glance(cal_year)

                self.create_page_index()

        # a single year
        else:
            cal_year = self.start_year

            # create webcalendar() calendar pages
            self.webcalendar(cal_year)

            # create "Year At A Glance" and
            # "One Day" calendar pages
            if self.fullyear:
                self.year_glance(cal_year)

            self.create_page_index()

    def create_page_index(self):
        """
        Create the page index called by the narrativeweb.
        """
        cal_fname = "cal/%s/index" % str(Today().get_year())
        output_file, sio = self.report.create_file(cal_fname)

        # Create page header
        # body has already been added to yearglance  already once
        result = self.write_header(cal_fname, cal=2)
        index, head, dummy_body, outerwrapper = result

        head += SCRIPT % self.ext

        # create blank line for stylesheets
        # write footer section
        footer = self.write_footer(None, cal=2)
        outerwrapper += (FULLCLEAR, footer)
        outerwrapper += "<script>currentmonth('.');</script>"

        # send calendar page to web output
        # and close the file
        self.xhtml_writer(index, output_file, sio, 0)


def _regular_surname(sex, name):
    """
    Returns a name string built from the components of the Name instance.
    """
    dummy_gender = sex
    surname = name.get_surname()
    suffix = name.get_suffix()
    if suffix:
        # TODO for Arabic, should the next line's comma be translated?
        surname = surname + ", " + suffix
    return surname


# Simple utility list to convert Gramps day-of-week numbering
# to calendar.firstweekday numbering
DOW_GRAMPS2ISO = [
    -1,
    calendar.SUNDAY,
    calendar.MONDAY,
    calendar.TUESDAY,
    calendar.WEDNESDAY,
    calendar.THURSDAY,
    calendar.FRIDAY,
    calendar.SATURDAY,
]


def get_marriage_event(db, family):
    """
    marriage_event will either be the marriage event or False
    """

    marriage_event = False
    for event_ref in family.get_event_ref_list():
        event = db.get_event_from_handle(event_ref.ref)
        if event.type.is_marriage():
            marriage_event = event
            break

    # return the marriage event or False to it caller
    return marriage_event


def get_divorce_event(db, family):
    """
    divorce will either be the divorce event or False
    """

    divorce_event = False
    for event_ref in family.get_event_ref_list():
        event = db.get_event_from_handle(event_ref.ref)
        if event.type.is_divorce():
            divorce_event = event
            break

    # return the divorce event or False to it caller
    return divorce_event


def get_first_day_of_month(year, month):
    """
    Compute the first day to display for this month.
    It can also be a day in the previous month.
    """

    # first day of the month
    current_date = datetime.date(year, month, 1)

    # monthinfo is filled using standard Python library
    # calendar.monthcalendar. It fills a list of 7-day-lists. The first day
    # of the 7-day-list is determined by calendar.firstweekday.
    monthinfo = calendar.monthcalendar(year, month)

    current_ord = current_date.toordinal() - monthinfo[0].count(0)
    return current_date, current_ord, monthinfo


def _has_webpage_extension(url):
    """
    determine if a filename has an extension or not...

    url = filename to be checked
    """
    return any(url.endswith(ext) for ext in _WEB_EXT)


def get_day_list(event_date, holiday_list, bday_anniv_list, rlocale=glocale):
    """
    Will fill day_list and return it to its caller: calendar_build()

    holiday_list -- list of holidays for event_date
    bday_anniv_list -- list of birthdays and anniversaries
        for event_date

    event_date -- date for this day_list

    'day_list' - a combination of both dictionaries to be able
        to create one day nyears, date, text, event --- are
        necessary for figuring the age or years married for
        each day being created...
    rlocale -- the locale to use
    """

    trans_text = rlocale.translation.sgettext
    # initialize day_list
    day_list = []

    ##################################################################
    # birthday/ anniversary on this day
    # Date.EMPTY signifies an incomplete date for an event. See add_day_item()
    bday_anniv_list = [
        (t, e, d, n, x) for t, e, d, n, x in bday_anniv_list if d != Date.EMPTY
    ]

    # number of years have to be at least zero
    bday_anniv_list = [
        (t, e, d, n, x)
        for t, e, d, n, x in bday_anniv_list
        if (event_date.get_year() - d.get_year()) >= 0
    ]

    # a holiday
    # zero will force holidays to be first in list
    nyears = 0

    for text, event, date, notused, notused in holiday_list:
        day_list.append((nyears, date, text, event, notused, notused))

    # birthday and anniversary list
    for text, event, date, age_at_death, dead_event_date in bday_anniv_list:
        # number of years married, ex: 10
        nyears = event_date.get_year() - date.get_year()

        # number of years for birthday, ex: 10 years
        age_str = event_date - date
        age_str = age_str.format(precision=1, as_age=False, dlocale=rlocale)

        symbols = Symbols()
        death_idx = config.get("utf8.death-symbol")
        death_symbol = symbols.get_death_symbol_for_char(death_idx)

        # a birthday
        if event == "Birthday":
            if age_at_death is not None:
                trans_date = trans_text("Died %(death_date)s.")
                translated_date = rlocale.get_date(dead_event_date)
                mess = trans_date % {"death_date": translated_date}
                age = ", <font size='+1' ><b>%s</b></font> <em>%s (%s)" % (
                    death_symbol,
                    mess,
                    age_at_death,
                )
            else:
                # Translators: expands to smth like "12 years old",
                # where "12 years" is already localized to your language
                age = ", <em>"
                date_y = date.get_year()
                trans_date = trans_text("Born %(birth_date)s.")
                old_date = trans_text("%s old")
                translated_date = rlocale.get_date(dead_event_date)
                age += old_date % (
                    str(age_str)
                    if (date_y != 0)
                    else trans_date % {"birth_date": translated_date}
                )
            txt_str = text + age + "</em>"

        # a death
        if event == "Death":
            txt_str = (
                text
                + ", "
                + death_symbol
                + " <em>"
                + (
                    trans_text("%s since death") % str(age_str)
                    if nyears
                    else trans_text("death")
                )
                + "</em>"
            )

        # an anniversary
        elif event == "Anniversary":
            if nyears == 0:
                txt_str = trans_text("%(couple)s, <em>wedding</em>") % {"couple": text}
            else:
                if age_at_death is not None:
                    age = "%s %s" % (trans_text("Married"), age_at_death)
                    txt_str = "%s, <em>%s" % (text, age)
                    if (
                        isinstance(dead_event_date, Date)
                        and dead_event_date.get_year() > 0
                    ):
                        txt_str += " (" + trans_text("Until") + " "
                        txt_str += rlocale.get_date(dead_event_date)
                        txt_str += ")</em>"
                    else:
                        txt_str += "</em>"
                else:
                    age = "<em>%s" % nyears
                    # Translators: leave all/any {...} untranslated
                    ngettext = rlocale.translation.ngettext
                    txt_str = ngettext(
                        "{couple}, {years} year anniversary",
                        "{couple}, {years} year anniversary",
                        nyears,
                    ).format(couple=text, years=age)
                    txt_str += "</em>"
            txt_str = Html("span", txt_str, class_="yearsmarried")

        day_list.append((nyears, date, txt_str, event, age_at_death, dead_event_date))

    # sort them based on number of years
    # holidays will always be on top of event list
    day_list = sorted(day_list, key=lambda x: (isinstance(x[0], str), x[0]))

    # return to its caller calendar_build()
    return day_list



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Base class for filter rules.
"""

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
import re

from ...errors import FilterError
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# enable logging for error handling
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".")


# -------------------------------------------------------------------------
#
# Rule
#
# -------------------------------------------------------------------------
class Rule:
    """Base rule class."""

    labels = []
    name = ""
    category = _("Miscellaneous filters")
    description = _("No description")
    allow_regex = False

    def __init__(self, arg, use_regex=False, use_case=False):
        self.list = []
        self.regex = []
        self.match_substring = self.__match_substring
        self.set_list(arg)
        self.use_regex = use_regex
        self.use_case = use_case
        self.nrprepare = 0

    def is_empty(self):
        return False

    def requestprepare(self, db, user):
        """
        Request that the prepare method of the rule is executed if needed

        Special: Custom Filters have fixed values, so only one instance needs to
        exists during a search. It is stored in a FilterStore, and initialized
        once.
        As filters are can be grouped in a group
        filter, we request a prepare. Only the first time prepare will be
        called
        """
        if self.nrprepare == 0:
            if self.use_regex:
                self.regex = [None] * len(self.labels)
                for index, label in enumerate(self.labels):
                    if self.list[index]:
                        try:
                            if self.use_case:
                                self.regex[index] = re.compile(self.list[index])
                            else:
                                self.regex[index] = re.compile(self.list[index], re.I)
                        except re.error:
                            self.regex[index] = re.compile("")
                self.match_substring = self.match_regex
            self.prepare(db, user)
        self.nrprepare += 1
        if self.nrprepare > 20:  # more references to a filter than expected
            raise FilterError(
                _("The filter definition contains a loop."),
                _(
                    "One rule references another which eventually"
                    " references the first."
                ),
            )

    def prepare(self, db, user):
        """prepare so the rule can be executed efficiently"""
        pass

    def requestreset(self):
        """
        Request that the reset method of the rule is executed if possible

        Special: Custom Filters have fixed values, so only one instance needs to
        exists during a search. It is stored in a FilterStore, and initialized
        once.
        As filters are can be grouped in a group
        filter, we request a reset. Only the last time reset will be
        called
        """
        self.nrprepare -= 1
        if self.nrprepare == 0:
            self.reset()

    def reset(self):
        """remove no longer needed memory"""
        pass

    def set_list(self, arg):
        """Store the values of this rule."""
        assert isinstance(arg, list) or arg is None, "Argument is not a list"
        if len(arg) != len(self.labels):
            LOG.warning(
                (
                    "Number of arguments does not match number of "
                    + "labels.\n   list:   %s\n   labels: %s"
                )
                % (arg, self.labels)
            )
        self.list = arg

    def values(self):
        """Return the values used by this rule."""
        return self.list

    def check(self):
        """Verify the number of rule values versus the number of rule labels."""
        return len(self.list) == len(self.labels)

    def apply(self, dummy_db, dummy_person):
        """Apply the rule to some database entry; must be overwritten."""
        return True

    def display_values(self):
        """Return the labels and values of this rule."""
        l_v = (
            '%s="%s"'
            % (
                _(
                    self.labels[index][0]
                    if isinstance(self.labels[index], tuple)
                    else self.labels[index]
                ),
                item,
            )
            for index, item in enumerate(self.list)
            if item
        )

        return ";".join(l_v)

    def __match_substring(self, param_index, str_var):
        """
        Return boolean indicating if database element represented by str_var
        matches filter element indicated by param_index using case insensitive
        string matching.
        """
        # make str_var unicode so that search for ü works
        # see issue 3188
        str_var = str(str_var)
        if self.list[param_index] and (
            str_var.upper().find(self.list[param_index].upper()) == -1
        ):
            return False
        else:
            return True

    def match_regex(self, param_index, str_var):
        """
        Return boolean indicating if database element represented by str_var
        matches filter element indicated by param_index using a regular
        expression search.
        """
        str_var = str(str_var)
        if self.list[param_index] and self.regex[param_index].search(str_var) is None:
            return False
        else:
            return True



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011 Nick Hall
# Copyright (C) 2016-     Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.editors import EditEvent
from gramps.gui.editors import EditPlace
from gramps.gui.listmodel import ListModel, NOSORT
from gramps.gen.plug import Gramplet
from gramps.gen.plug.report.utils import find_spouse
from gramps.gui.dbguielement import DbGUIElement
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.datehandler import get_date
from gramps.gen.utils.place import conv_lat_lon
from gramps.gen.utils.db import (
    get_participant_from_event,
    get_birth_or_fallback,
    get_marriage_or_fallback,
)
from gramps.gen.errors import WindowActiveError
from gramps.gen.config import config
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class GeoEvents(Gramplet, DbGUIElement):
    def __init__(self, gui, nav_group=0):
        Gramplet.__init__(self, gui, nav_group)
        DbGUIElement.__init__(self, self.dbstate.db)

    """
    Displays the events for a person or family.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.gui.WIDGET.show()

    def _connect_db_signals(self):
        """
        called on init of DbGUIElement, connect to db as required.
        """
        self.callman.register_callbacks({"event-update": self.changed})
        self.callman.connect_all(keys=["event"])

    def changed(self, handle):
        """
        Called when a registered event is updated.
        """
        self.update()

    def build_gui(self):
        """
        Build the GUI interface.
        """
        tip = _(
            "Right-click on a row to edit the selected event" " or the related place."
        )
        self.set_tooltip(tip)
        top = Gtk.TreeView()
        top.set_hover_selection(True)
        titles = [
            (
                "",
                NOSORT,
                50,
            ),
            (_("Type"), 1, 100),
            (_("Description"), 2, 250),
            (_("Date"), 3, 160),
            ("", NOSORT, 50),
            (_("Place"), 4, 300),
            (_("Id"), 5, 80),
            (_("Latitude"), 6, 130),
            (_("Longitude"), 7, 130),
        ]
        self.model = ListModel(top, titles, right_click=self.menu_edit)
        return top

    def add_event_ref(self, event_ref, spouse=None, name=""):
        """
        Add an event to the model.
        """
        self.callman.register_handles({"event": [event_ref.ref]})
        event = self.dbstate.db.get_event_from_handle(event_ref.ref)
        event_date = get_date(event)
        event_sort = "%012d" % event.get_date_object().get_sort_value()
        place_name = place_displayer.display_event(self.dbstate.db, event)
        place_handle = event.get_place_handle()
        place_id = latitude = longitude = ""
        if place_handle:
            plc = self.dbstate.db.get_place_from_handle(place_handle)
            if plc:
                place_id = plc.get_gramps_id()
                latitude = plc.get_latitude()
                longitude = plc.get_longitude()
                latitude, longitude = conv_lat_lon(latitude, longitude, "D.D8")

        description = event.get_description()
        if description == "":
            description = name

        self.model.add(
            (
                event.get_handle(),
                str(event.get_type()),
                description,
                event_date,
                event_sort,
                place_name,
                place_id,
                latitude,
                longitude,
            )
        )

    def menu_edit(self, treeview, event):
        """
        Show a menu to select either Edit the selected event or
        the Place related to this event.
        """
        self.menu = Gtk.Menu()
        menu = self.menu
        title = _("Edit the event")
        add_item = Gtk.MenuItem(label=title)
        add_item.connect("activate", self.edit_event, treeview)
        add_item.show()
        menu.append(add_item)
        title = _("Edit the place")
        add_item = Gtk.MenuItem(label=title)
        add_item.connect("activate", self.edit_place, treeview)
        add_item.show()
        menu.append(add_item)
        menu.show()
        menu.popup_at_pointer(event)

    def edit_place(self, menuitem, treeview):
        """
        Edit the place related to the selected event.
        """
        model, iter_ = treeview.get_selection().get_selected()
        if iter_:
            handle = model.get_value(iter_, 0)
            try:
                event = self.dbstate.db.get_event_from_handle(handle)
                place_handle = event.get_place_handle()
                place_id = latitude = longitude = ""
                if place_handle:
                    plc = self.dbstate.db.get_place_from_handle(place_handle)
                    EditPlace(self.dbstate, self.uistate, [], plc)
            except WindowActiveError:
                pass

    def edit_event(self, menuitem, treeview):
        """
        Edit the selected event.
        """
        model, iter_ = treeview.get_selection().get_selected()
        if iter_:
            handle = model.get_value(iter_, 0)
            try:
                event = self.dbstate.db.get_event_from_handle(handle)
                EditEvent(self.dbstate, self.uistate, [], event)
            except WindowActiveError:
                pass


class GeoPersonEvents(GeoEvents):
    """
    Displays the events for a person.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "person-update", self.update)

    def active_changed(self, handle):
        self.update()

    def update_has_data(self):
        active_handle = self.get_active("Person")
        active = None
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
        self.set_has_data(self.get_has_data(active))

    def get_has_data(self, active_person):
        """
        Return True if the gramplet has data, else return False.
        """
        if active_person:
            if active_person.get_event_ref_list():
                return True
            for family_handle in active_person.get_family_handle_list():
                family = self.dbstate.db.get_family_from_handle(family_handle)
                if family:
                    for event_ref in family.get_event_ref_list():
                        return True
        return False

    def main(self):  # return false finishes
        active_handle = self.get_active("Person")

        self.model.clear()
        self.callman.unregister_all()
        if active_handle:
            self.display_person(active_handle)
        else:
            self.set_has_data(False)

    def display_person(self, active_handle):
        """
        Display the events for the active person.
        """
        active_person = self.dbstate.db.get_person_from_handle(active_handle)
        if active_person:
            for event_ref in active_person.get_event_ref_list():
                surname = name_displayer.display(active_person)
                self.add_event_ref(event_ref, name=surname)
            for family_handle in active_person.get_family_handle_list():
                family = self.dbstate.db.get_family_from_handle(family_handle)
                self.display_family(family, active_person)
        self.set_has_data(self.model.count > 0)

    def display_family(self, family, active_person):
        """
        Display the events for the given family.
        """
        spouse_handle = find_spouse(active_person, family)
        if spouse_handle:
            spouse = self.dbstate.db.get_person_from_handle(spouse_handle)
        else:
            spouse = None
        if family:
            for event_ref in family.get_event_ref_list():
                self.add_event_ref(event_ref, spouse)


class GeoFamilyEvents(GeoEvents):
    """
    Displays the events for a family.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "family-update", self.update)
        self.connect_signal("Family", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Family")
        active = None
        if active_handle:
            active = self.dbstate.db.get_family_from_handle(active_handle)
        self.set_has_data(self.get_has_data(active))

    def get_has_data(self, active_family):
        """
        Return True if the gramplet has data, else return False.
        """
        if active_family:
            for event_ref in active_family.get_event_ref_list():
                return True
        return False

    def main(self):  # return false finishes
        active_handle = self.get_active("Family")

        self.model.clear()
        self.callman.unregister_all()
        if active_handle:
            self.display_family(active_handle)
        else:
            self.set_has_data(False)

    def display_family(self, active_handle):
        """
        Display the events for the active family.
        """
        active_family = self.dbstate.db.get_family_from_handle(active_handle)
        for event_ref in active_family.get_event_ref_list():
            self.add_event_ref(event_ref)
        self.set_has_data(self.model.count > 0)
        # for father, add all events
        father_handle = active_family.get_father_handle()
        if father_handle:
            father = self.dbstate.db.get_person_from_handle(father_handle)
            if father:
                for event_ref in father.get_event_ref_list():
                    surname = name_displayer.display(father)
                    self.add_event_ref(event_ref, name=surname)
        # for mother, add all events
        mother_handle = active_family.get_mother_handle()
        if mother_handle:
            mother = self.dbstate.db.get_person_from_handle(mother_handle)
            if mother:
                for event_ref in mother.get_event_ref_list():
                    surname = name_displayer.display(mother)
                    self.add_event_ref(event_ref, name=surname)
        # for each child, add all events
        child_ref_list = active_family.get_child_ref_list()
        if child_ref_list:
            for chld_ref in child_ref_list:
                if chld_ref:
                    child = self.dbstate.db.get_person_from_handle(chld_ref.ref)
                    if child:
                        for event_ref in child.get_event_ref_list():
                            surname = name_displayer.display(child)
                            self.add_event_ref(event_ref, name=surname)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#               2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import abc

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository.Gio import SimpleActionGroup

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ..managedwindow import ManagedWindow
from gramps.gen.datehandler import displayer, parser
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.config import config
from ..utils import is_right_click
from ..display import display_help
from ..dialog import SaveDialog
from gramps.gen.lib import PrimaryObject
from ..dbguielement import DbGUIElement
from ..uimanager import ActionGroup


class EditPrimary(ManagedWindow, DbGUIElement, metaclass=abc.ABCMeta):
    QR_CATEGORY = -1

    def __init__(
        self,
        state,
        uistate,
        track,
        obj,
        get_from_handle,
        get_from_gramps_id,
        callback=None,
    ):
        """
        Create an edit window.

        Associate a person with the window.

        """
        self.dp = parser
        self.dd = displayer
        self.name_displayer = name_displayer
        self.obj = obj
        self.dbstate = state
        self.uistate = uistate
        self.db = state.db
        self.callback = callback
        self.ok_button = None
        self.get_from_handle = get_from_handle
        self.get_from_gramps_id = get_from_gramps_id
        self.contexteventbox = None
        self.__tabs = []
        self.action_group = None

        ManagedWindow.__init__(self, uistate, track, obj)
        DbGUIElement.__init__(self, self.db)

        self.original = None
        if self.obj.handle:
            self.original = self.get_from_handle(self.obj.handle)

        self._local_init()
        # self.set_size() is called by self._local_init()'s self.setup_configs
        self._create_tabbed_pages()
        self._setup_fields()
        self._connect_signals()
        # if the database is changed, all info shown is invalid and the window
        # should close
        self.dbstate_connect_key = self.dbstate.connect(
            "database-changed", self._do_close
        )
        self.show()
        self._post_init()

    def _local_init(self):
        """
        Derived class should do any pre-window initalization in this task.
        """
        pass

    def _post_init(self):
        """
        Derived class should do any post-window initalization in this task.
        """
        pass

    def _setup_fields(self):
        pass

    def _create_tabbed_pages(self):
        pass

    def _connect_signals(self):
        pass

    def build_window_key(self, obj):
        if obj and obj.get_handle():
            return obj.get_handle()
        else:
            return id(self)

    def _setup_notebook_tabs(self, notebook):
        for child in notebook.get_children():
            label = notebook.get_tab_label(child)
            page_no = notebook.page_num(child)
            label.drag_dest_set(0, [], 0)
            label.connect("drag_motion", self._switch_page_on_dnd, notebook, page_no)
            child.set_parent_notebook(notebook)

        notebook.connect("key-press-event", self.key_pressed, notebook)

    def key_pressed(self, obj, event, notebook):
        """
        Handles the key being pressed on the notebook, pass to key press of
        current page.
        """
        pag = notebook.get_current_page()
        if not pag == -1:
            notebook.get_nth_page(pag).key_pressed(obj, event)

    def _switch_page_on_dnd(self, widget, context, x, y, time, notebook, page_no):
        if notebook.get_current_page() != page_no:
            notebook.set_current_page(page_no)

    def _add_tab(self, notebook, page):
        self.__tabs.append(page)
        notebook.insert_page(page, page.get_tab_widget(), -1)
        page.label.set_use_underline(True)
        return page

    def _cleanup_on_exit(self):
        """Unset all things that can block garbage collection.
        Finalize rest
        """
        for tab in self.__tabs:
            if hasattr(tab, "_cleanup_on_exit"):
                tab._cleanup_on_exit()
        self.__tabs = None

    def object_is_empty(self):
        return self.obj.serialize()[1:] == self.empty_object().serialize()[1:]

    def define_ok_button(self, button, function):
        self.ok_button = button
        button.connect("clicked", function)
        button.set_sensitive(not self.db.readonly)

    def define_cancel_button(self, button):
        button.connect("clicked", self.close)

    def define_help_button(self, button, webpage="", section=""):
        button.connect("clicked", lambda x: display_help(webpage, section))

    def _do_close(self, *obj):
        self._cleanup_db_connects()
        self.dbstate.disconnect(self.dbstate_connect_key)
        self._cleanup_connects()
        self._cleanup_on_exit()
        if self.action_group:
            self.uistate.uimanager.remove_action_group(self.action_group)
        self.get_from_handle = None
        self.get_from_gramps_id = None
        ManagedWindow.close(self)
        self.dbstate = None
        self.uistate = None
        self.db = None

    def _cleanup_db_connects(self):
        """
        All connects that happened to signals of the db must be removed on
        closed. This implies two things:
        1. The connects on the main view must be disconnected
        2. Connects done in subelements must be disconnected
        """
        # cleanup callbackmanager of this editor
        self._cleanup_callbacks()
        for tab in self.__tabs:
            if hasattr(tab, "callman"):
                tab._cleanup_callbacks()

    def _cleanup_connects(self):
        """
        Connects to interface elements to things outside the element should be
        removed before destroying the interface
        """
        self._cleanup_local_connects()
        for tab in [
            tab for tab in self.__tabs if hasattr(tab, "_cleanup_local_connects")
        ]:
            tab._cleanup_local_connects()

    def _cleanup_local_connects(self):
        """
        Connects to interface elements to things outside the element should be
        removed before destroying the interface. This methods cleans connects
        of the main interface, not of the displaytabs.
        """
        pass

    def check_for_close(self, handles):
        """
        Callback method for delete signals.
        If there is a delete signal of the primary object we are editing, the
        editor (and all child windows spawned) should be closed
        """
        if self.obj.get_handle() in handles:
            self._do_close()

    def close(self, *obj):
        """If the data has changed, give the user a chance to cancel
        the close window"""
        if not config.get("interface.dont-ask") and self.data_has_changed():
            SaveDialog(
                _("Save Changes?"),
                _(
                    "If you close without saving, the changes you "
                    "have made will be lost"
                ),
                self._do_close,
                self.save,
                parent=self.window,
            )
            return True
        else:
            self._do_close()
            return False

    @abc.abstractmethod
    def empty_object(self):
        """empty_object should be overridden in child class"""

    def data_has_changed(self):
        if self.db.readonly:
            return False
        if self.original:
            cmp_obj = self.original
        else:
            cmp_obj = self.empty_object()
        return cmp_obj.serialize()[1:] != self.obj.serialize()[1:]

    def save(self, *obj):
        """Save changes and close. Inheriting classes must implement this"""
        self.close()

    def set_contexteventbox(self, eventbox):
        """Set the contextbox that grabs button presses if not grabbed
        by overlying widgets.
        """
        self.contexteventbox = eventbox
        self.contexteventbox.connect(
            "button-press-event", self._contextmenu_button_press
        )

    def _contextmenu_button_press(self, obj, event):
        """
        Button press event that is caught when a mousebutton has been
        pressed while on contexteventbox
        It opens a context menu with possible actions
        """
        if is_right_click(event):
            if self.obj.get_handle() == 0:
                return False

            # build the possible popup menu
            menu_model = self._build_popup_ui()
            if not menu_model:
                return False
            # set or unset sensitivity in popup
            self._post_build_popup_ui()

            menu = Gtk.Menu.new_from_model(menu_model)
            menu.attach_to_widget(obj, None)
            menu.show_all()
            menu.popup_at_pointer(event)
            return True
        return False

    def _build_popup_ui(self):
        """
        Create actions and ui of context menu
        If you don't need a popup, override this and return None
        """
        from ..plug.quick import create_quickreport_menu

        prefix = str(id(self))
        # get custom ui and actions
        (ui_top, actions) = self._top_contextmenu(prefix)
        # see which quick reports are available now:
        ui_qr = ""
        if self.QR_CATEGORY > -1:
            (ui_qr, reportactions) = create_quickreport_menu(
                self.QR_CATEGORY,
                self.dbstate,
                self.uistate,
                self.obj,
                prefix,
                track=self.track,
            )
            actions.extend(reportactions)

        popupui = (
            """<?xml version="1.0" encoding="UTF-8"?>
            <interface>
            <menu id="Popup">"""
            + ui_top
            + """
              <section>
            """
            + ui_qr
            + """
              </section>
            </menu>
            </interface>"""
        )

        builder = Gtk.Builder.new_from_string(popupui, -1)

        self.action_group = ActionGroup("EditPopup" + prefix, actions, prefix)
        act_grp = SimpleActionGroup()
        self.window.insert_action_group(prefix, act_grp)
        self.window.set_application(self.uistate.uimanager.app)
        self.uistate.uimanager.insert_action_group(self.action_group, act_grp)
        return builder.get_object("Popup")

    def _top_contextmenu(self, prefix):
        """
        Derived class can create a ui with menuitems and corresponding list of
        actiongroups
        """
        return "", []

    def _post_build_popup_ui(self):
        """
        Derived class should do extra actions here on the menu
        """
        pass

    def _uses_duplicate_id(self):
        """
        Check whether a changed or added Gramps ID already exists in the DB.

        Return True if a duplicate Gramps ID has been detected.

        """
        idval = self.obj.get_gramps_id()
        existing = self.get_from_gramps_id(idval)
        if existing:
            if existing.get_handle() == self.obj.get_handle():
                return (False, 0)
            else:
                return (True, idval)
        else:
            return (False, 0)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004  Martin Hawlisch
# Copyright (C) 2005-2008  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Export Persons to vCard (RFC 2426)."

# -------------------------------------------------------------------------
#
# Standard Python Modules
#
# -------------------------------------------------------------------------
import sys
from textwrap import TextWrapper

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging
from collections import abc

log = logging.getLogger(".ExportVCard")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.const import PROGRAM_NAME
from gramps.version import VERSION
from gramps.gen.lib import Date, Person
from gramps.gen.lib.urltype import UrlType
from gramps.gui.plug.export import WriterOptionBox
from gramps.gen.lib.eventtype import EventType
from gramps.gen.display.name import displayer as _nd
from gramps.gen.plug.utils import OpenFileOrStdout


# -------------------------------------------------------------------------
#
# Support Functions
#
# -------------------------------------------------------------------------
def exportData(database, filename, user, option_box=None):
    """Function called by Gramps to export data on persons in VCard format."""
    cardw = VCardWriter(database, filename, option_box, user)
    try:
        cardw.export_data()
    except EnvironmentError as msg:
        user.notify_error(_("Could not create %s") % filename, str(msg))
        return False
    except:
        # Export shouldn't bring Gramps down.
        user.notify_error(_("Could not create %s") % filename)
        return False
    return True


# -------------------------------------------------------------------------
#
# VCardWriter class
#
# -------------------------------------------------------------------------
class VCardWriter:
    """Class to create a file with data in VCard format."""

    LINELENGTH = 73  # unclear if the 75 chars of spec includes \r\n.
    ESCAPE_CHAR = "\\"
    TOBE_ESCAPED = ["\\", ",", ";"]  # order is important
    LINE_CONTINUATION = [" ", "\t"]

    @staticmethod
    def esc(data):
        """Escape the special chars of the VCard protocol."""
        if isinstance(data, str):
            for char in VCardWriter.TOBE_ESCAPED:
                data = data.replace(char, VCardWriter.ESCAPE_CHAR + char)
            return data
        elif type(data) == type([]):
            return list(map(VCardWriter.esc, data))
        elif type(data) == type(()):
            return tuple(map(VCardWriter.esc, data))
        else:
            raise TypeError(
                "VCard escaping is not implemented for "
                "data type %s." % str(type(data))
            )

    def __init__(self, database, filename, option_box=None, user=None):
        self.db = database
        self.filename = filename
        self.user = user
        self.filehandle = None
        self.option_box = option_box
        if isinstance(self.user.callback, abc.Callable):  # is really callable
            self.update = self.update_real
        else:
            self.update = self.update_empty

        if option_box:
            self.option_box.parse_options()
            self.db = option_box.get_filtered_database(self.db)

        self.txtwrp = TextWrapper(
            width=self.LINELENGTH,
            expand_tabs=False,
            replace_whitespace=False,
            drop_whitespace=False,
            subsequent_indent=self.LINE_CONTINUATION[0],
        )
        self.count = 0
        self.total = 0

    def update_empty(self):
        """Progress can't be reported."""
        pass

    def update_real(self):
        """Report progress."""
        self.count += 1
        newval = int(100 * self.count / self.total)
        if newval != self.oldval:
            self.user.callback(newval)
            self.oldval = newval

    def writeln(self, text):
        """
        Write a property of the VCard to file.

        Can't cope with nested VCards, section 2.4.2 of RFC 2426.
        """
        self.filehandle.write(
            "%s\r\n" % "\r\n".join([line for line in self.txtwrp.wrap(text)])
        )

    def export_data(self):
        """Open the file and loop over everyone too write their VCards."""
        with OpenFileOrStdout(
            self.filename, encoding="utf-8", errors="strict", newline=""
        ) as self.filehandle:
            if self.filehandle:
                self.count = 0
                self.oldval = 0
                self.total = self.db.get_number_of_people()
                for key in sorted(list(self.db.iter_person_handles())):
                    self.write_person(key)
                    self.update()
        return True

    def write_person(self, person_handle):
        """Create a VCard for the specified person."""
        person = self.db.get_person_from_handle(person_handle)
        if person:
            self.write_header()
            prname = person.get_primary_name()
            self.write_formatted_name(prname)
            self.write_name(prname)
            self.write_sortstring(prname)
            self.write_nicknames(person, prname)
            self.write_gender(person)
            self.write_birthdate(person)
            self.write_addresses(person)
            self.write_urls(person)
            self.write_occupation(person)
            self.write_footer()

    def write_header(self):
        """Write the opening lines of a VCard."""
        self.writeln("BEGIN:VCARD")
        self.writeln("VERSION:3.0")
        self.writeln("PRODID:-//Gramps//NONSGML %s %s//EN" % (PROGRAM_NAME, VERSION))

    def write_footer(self):
        """Write the closing lines of a VCard."""
        self.writeln("END:VCARD")
        self.writeln("")

    def write_formatted_name(self, prname):
        """Write the compulsory FN property of VCard."""
        regular_name = prname.get_regular_name().strip()
        title = prname.get_title()
        if title:
            regular_name = "%s %s" % (title, regular_name)
        self.writeln("FN:%s" % self.esc(regular_name))

    def write_name(self, prname):
        """Write the compulsory N property of a VCard."""
        family_name = ""
        given_name = ""
        additional_names = ""
        hon_prefix = ""
        suffix = ""

        primary_surname = prname.get_primary_surname()
        surname_list = prname.get_surname_list()
        if not surname_list[0].get_primary():
            surname_list.remove(primary_surname)
            surname_list.insert(0, primary_surname)
        family_name = ",".join(
            self.esc(
                [
                    (
                        "%s %s %s"
                        % (
                            surname.get_prefix(),
                            surname.get_surname(),
                            surname.get_connector(),
                        )
                    ).strip()
                    for surname in surname_list
                ]
            )
        )

        call_name = prname.get_call_name()
        if call_name:
            given_name = self.esc(call_name)
            additional_name_list = prname.get_first_name().split()
            if call_name in additional_name_list:
                additional_name_list.remove(call_name)
            additional_names = ",".join(self.esc(additional_name_list))
        else:
            name_list = prname.get_first_name().split()
            if len(name_list) > 0:
                given_name = self.esc(name_list[0])
                if len(name_list) > 1:
                    additional_names = ",".join(self.esc(name_list[1:]))
        # Alternate names are ignored because names just don't add up:
        # if one name is Jean and an alternate is Paul then you can't
        # conclude the Jean Paul is also an alternate name of that person.

        # Assume all titles/suffixes that apply are present in primary name.
        hon_prefix = ",".join(self.esc(prname.get_title().split()))
        suffix = ",".join(self.esc(prname.get_suffix().split()))

        self.writeln(
            "N:%s;%s;%s;%s;%s"
            % (family_name, given_name, additional_names, hon_prefix, suffix)
        )

    def write_sortstring(self, prname):
        """Write the SORT-STRING property of a VCard."""
        # TODO only add sort-string if needed
        self.writeln("SORT-STRING:%s" % self.esc(_nd.sort_string(prname)))

    def write_nicknames(self, person, prname):
        """Write the NICKNAME property of a VCard."""
        nicknames = [
            x.get_nick_name() for x in person.get_alternate_names() if x.get_nick_name()
        ]
        if prname.get_nick_name():
            nicknames.insert(0, prname.get_nick_name())
        if len(nicknames) > 0:
            self.writeln("NICKNAME:%s" % (",".join(self.esc(nicknames))))

    def write_gender(self, person):
        """Write the X-GENDER property of a VCard (X- dropped in 4.0, we're at 3.0)."""
        gender = person.get_gender()
        gender_value = ""
        if gender == Person.MALE:
            gender_value = "Male"
        elif gender == Person.FEMALE:
            gender_value = "Female"
        elif gender == Person.OTHER:
            gender_value = "Other"
        log.info("gender: %s -> %s" % (gender, gender_value))
        if gender_value:
            self.writeln("X-GENDER:%s" % (gender_value))

    def write_birthdate(self, person):
        """Write the BDAY property of a VCard."""
        birth_ref = person.get_birth_ref()
        if birth_ref:
            birth = self.db.get_event_from_handle(birth_ref.ref)
            if birth:
                b_date = birth.get_date_object()
                mod = b_date.get_modifier()
                if (
                    mod != Date.MOD_TEXTONLY
                    and not b_date.is_empty()
                    and not mod == Date.MOD_SPAN
                    and not mod == Date.MOD_FROM
                    and not mod == Date.MOD_TO
                    and not mod == Date.MOD_RANGE
                ):
                    (day, month, year, slash) = b_date.get_start_date()
                    if day > 0 and month > 0 and year > 0:
                        self.writeln("BDAY:%s-%02d-%02d" % (year, month, day))

    def write_addresses(self, person):
        """Write ADR and TEL properties of a VCard."""
        address_list = person.get_address_list()
        for address in address_list:
            postbox = ""
            ext = ""
            street = address.get_street()
            city = address.get_city()
            state = address.get_state()
            zipcode = address.get_postal_code()
            country = address.get_country()
            if street or city or state or zipcode or country:
                self.writeln(
                    "ADR:%s;%s;%s;%s;%s;%s;%s"
                    % self.esc((postbox, ext, street, city, state, zipcode, country))
                )

            phone = address.get_phone()
            if phone:
                self.writeln("TEL:%s" % phone)

    def write_urls(self, person):
        """Write URL and EMAIL properties of a VCard."""
        url_list = person.get_url_list()
        for url in url_list:
            href = url.get_path()
            if href:
                if url.get_type() == UrlType(UrlType.EMAIL):
                    if href.startswith("mailto:"):
                        href = href[len("mailto:") :]
                    self.writeln("EMAIL:%s" % self.esc(href))
                else:
                    self.writeln("URL:%s" % self.esc(href))

    def write_occupation(self, person):
        """
        Write ROLE property of a VCard.

        Use the most recent occupation event.
        """
        event_refs = person.get_primary_event_ref_list()
        events = [
            event
            for event in [self.db.get_event_from_handle(ref.ref) for ref in event_refs]
            if event.get_type() == EventType(EventType.OCCUPATION)
        ]
        if len(events) > 0:
            events.sort(key=lambda x: x.get_date_object())
            occupation = events[-1].get_description()
            if occupation:
                self.writeln("ROLE:%s" % occupation)



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011-2016       Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import os
from math import pi, sin, atanh, floor

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import time
import logging

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------
from gramps.plugins.lib.maps import constants
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.const import VERSION_DIR
from gramps.gen.config import config
from gramps.gui.dialog import ErrorDialog
from gramps.gen.constfunc import get_env_var
from .dummylayer import DummyLayer
from .dummynogps import DummyMapNoGpsPoint
from .selectionlayer import SelectionLayer
from .lifewaylayer import LifeWayLayer
from .markerlayer import MarkerLayer
from .datelayer import DateLayer
from .messagelayer import MessageLayer
from .kmllayer import KmlLayer

# -------------------------------------------------------------------------
#
# OsmGps
#
# -------------------------------------------------------------------------

_LOG = logging.getLogger("maps.osmgps")
_ = glocale.translation.sgettext

try:
    import gi

    gi.require_version("OsmGpsMap", "1.0")
    from gi.repository import OsmGpsMap as osmgpsmap
except:
    raise


def lon2pixel(zoom, longitude, size):
    """
    pixel_x = (2^zoom * size * longitude) / 2PI + (2^zoom * size) / 2
    """
    value = ((longitude * size * (2**zoom)) / (2 * pi)) + ((2**zoom) * size / 2)
    return int(value)


def lat2pixel(zoom, latitude, size):
    """
    http://manialabs.wordpress.com/2013/01/26/converting-latitude-and-longitude-to-map-tile-in-mercator-projection/

    pixel_y = -(2^zoom * size * lat_m) / 2PI + (2^zoom * size) / 2
    """
    lat_m = atanh(sin(latitude))
    value = -((lat_m * size * (2**zoom)) / (2 * pi)) + ((2**zoom) * (size / 2))
    return int(value)


class OsmGps:
    """
    This class is used to create a map
    """

    def __init__(self, uistate):
        """
        Initialize the map
        """
        self.vbox = None
        self.cross_map = None
        self.osm = None
        self.show_tooltips = True
        self.zone_selection = False
        self.selection_layer = None
        self.lifeway_layer = None
        self.marker_layer = None
        self.date_layer = None
        self.message_layer = None
        self.kml_layer = None
        self.context_id = 0
        self.begin_selection = None
        self.end_selection = None
        self.current_map = None
        self.places_found = None
        self.uistate = uistate
        self.zoom = config.get("geography.zoom")

    def build_widget(self):
        """
        create the vbox
        """
        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        cache_path = config.get("geography.path")
        if not os.path.isdir(cache_path):
            try:
                os.makedirs(cache_path, 0o755)  # create dir like mkdir -p
            except:
                ErrorDialog(
                    _("Can't create " "tile cache directory %s") % cache_path,
                    parent=self.uistate.window,
                )
                gini = os.path.join(VERSION_DIR, "gramps.ini")
                ErrorDialog(
                    _(
                        "You must verify and change the tile cache"
                        "\n..."
                        "\n[geography]"
                        "\n..."
                        "\npath='bad/path'"
                        "\n..."
                        "\nin the gramps.ini file :\n%s"
                        "\n\nBefore changing the gramps.ini file, "
                        "you need to close Gramps"
                        "\n\nThe following errors are expected"
                    )
                    % gini,
                    parent=self.uistate.window,
                )
                return None

        self.change_map(None, config.get("geography.map_service"))
        return self.vbox

    def change_map(self, obj, map_type):
        """
        Change the current map
        """
        # if map_type == constants.PERSONAL:
        #     map_source = config.get("geography.pers# onal-map")
        #     if map_source == "":
        #         return
        #     name = constants.TILES_PATH[map_type]
        #     self.change_new_map(name, map_source)
        #     config.set("geography.map_service", map_type)
        #     self.current_map = map_type
        #     return
        if obj is not None:
            self.osm.layer_remove_all()
            self.osm.image_remove_all()
            self.vbox.remove(self.osm)
            self.osm.destroy()
        tiles_path = os.path.join(
            config.get("geography.path"), constants.TILES_PATH[map_type]
        )
        if not os.path.isdir(tiles_path):
            try:
                os.makedirs(tiles_path, 0o755)  # create dir like mkdir -p
            except:
                ErrorDialog(
                    _("Can't create " "tiles cache directory for '%s'.")
                    % constants.MAP_TITLE[map_type],
                    parent=self.uistate.window,
                )
        config.set("geography.map_service", map_type)
        self.current_map = map_type
        http_proxy = get_env_var("http_proxy")
        if 0:
            self.osm = DummyMapNoGpsPoint()
        else:
            # if map_type == constants.PERSONAL:
            #     self.osm = osmgpsmap.Map(repo_uri=map_source)
            # else:
            #     self.osm = osmgpsmap.Map(map_source=constants.MAP_TYPE[map_type])
            self.osm = osmgpsmap.Map(map_source=constants.MAP_TYPE[map_type])
            if http_proxy:
                self.osm.set_property("proxy_uri", http_proxy)
            self.osm.set_property("tile_cache", tiles_path)
        self.osm.props.tile_cache = osmgpsmap.MAP_CACHE_AUTO
        current_map = osmgpsmap.MapOsd(show_dpad=False, show_zoom=True)
        self.end_selection = None
        self.osm.layer_add(current_map)
        self.osm.layer_add(DummyLayer())
        self.selection_layer = self.add_selection_layer()
        self.kml_layer = self.add_kml_layer()
        self.lifeway_layer = self.add_lifeway_layer()
        self.marker_layer = self.add_marker_layer()
        self.date_layer = self.add_date_layer()
        self.message_layer = self.add_message_layer()
        self.cross_map = osmgpsmap.MapOsd(show_crosshair=False)
        self.set_crosshair(config.get("geography.show_cross"))
        self.osm.set_center_and_zoom(
            config.get("geography.center-lat"),
            config.get("geography.center-lon"),
            config.get("geography.zoom"),
        )

        self.osm.connect("button_release_event", self.map_clicked)
        self.osm.connect("button_press_event", self.map_clicked)
        self.osm.connect("motion-notify-event", self.motion_event)
        self.osm.connect("changed", self.zoom_changed)
        self.update_shortcuts(True)
        self.osm.show()
        self.vbox.pack_start(self.osm, True, True, 0)
        self.goto_handle(handle=None)

    def change_new_map(self, name, map_source):
        """
        Change the current map with a new provider
        This map is not supported by osm-gps-map

        name       : the name of the provider
        map_source : the url to search for tiles
        """
        try:
            self.osm.layer_remove_all()
            self.osm.image_remove_all()
            self.vbox.remove(self.osm)
            self.osm.destroy()
        except:
            pass
        tiles_path = os.path.join(config.get("geography.path"), name)
        if not os.path.isdir(tiles_path):
            try:
                os.makedirs(tiles_path, 0o755)  # create dir like mkdir -p
            except:
                ErrorDialog(
                    _("Can't create " "tiles cache directory for '%s'.")
                    % constants.MAP_TITLE[self.current_map],
                    parent=self.uistate.window,
                )
        http_proxy = get_env_var("http_proxy")
        if 0:
            self.osm = DummyMapNoGpsPoint()
        else:
            map_type = int(config.get("geography.map_service"))
            # if map_type == constants.PERSONAL:
            #     self.osm = osmgpsmap.Map(repo_uri=map_source)
            # else:
            #     self.osm = osmgpsmap.Map(map_source=constants.MAP_TYPE[map_type])
            self.osm = osmgpsmap.Map(map_source=constants.MAP_TYPE[map_type])
            if http_proxy:
                self.osm.set_property("proxy_uri", http_proxy)
            self.osm.set_property("tile_cache", tiles_path)
        self.osm.props.tile_cache = osmgpsmap.MAP_CACHE_AUTO
        current_map = osmgpsmap.MapOsd(show_dpad=False, show_zoom=True)
        self.end_selection = None
        self.osm.layer_add(current_map)
        self.osm.layer_add(DummyLayer())
        self.selection_layer = self.add_selection_layer()
        self.kml_layer = self.add_kml_layer()
        self.lifeway_layer = self.add_lifeway_layer()
        self.marker_layer = self.add_marker_layer()
        self.date_layer = self.add_date_layer()
        self.message_layer = self.add_message_layer()
        self.cross_map = osmgpsmap.MapOsd(show_crosshair=False)
        self.set_crosshair(config.get("geography.show_cross"))
        self.osm.set_center_and_zoom(
            config.get("geography.center-lat"),
            config.get("geography.center-lon"),
            config.get("geography.zoom"),
        )

        self.osm.connect("button_release_event", self.map_clicked)
        self.osm.connect("button_press_event", self.map_clicked)
        self.osm.connect("motion-notify-event", self.motion_event)
        self.osm.connect("changed", self.zoom_changed)
        self.update_shortcuts(True)
        self.osm.show()
        self.vbox.pack_start(self.osm, True, True, 0)
        self.goto_handle(handle=None)

    def reload_tiles(self):
        """
        We need to reload all visible tiles for the current map
        """
        map_idx = config.get("geography.map_service")
        map_name = constants.TILES_PATH[map_idx]
        map_path = os.path.join(config.get("geography.path"), map_name)
        # get the top left corner and bottom right corner
        bbox = self.osm.get_bbox()
        pt1 = bbox[0]
        pt2 = bbox[1]
        self.zoom = config.get("geography.zoom")
        tile_size = float(256)
        # if map_idx != constants.PERSONAL:
        #     # get the file extension depending on the map provider
        #     img_format = self.osm.source_get_image_format(map_idx)
        # else:
        #     filename = config.get("geography.personal-map")
        #     img_format = os.path.splitext(filename)[1]
        # get the file extension depending on the map provider
        img_format = self.osm.source_get_image_format(map_idx)
        # calculate the number of images to download in rows and columns
        pt1_x = floor(lon2pixel(self.zoom, pt1.rlon, tile_size) / tile_size)
        pt1_y = floor(lat2pixel(self.zoom, pt1.rlat, tile_size) / tile_size)
        pt2_x = floor(lon2pixel(self.zoom, pt2.rlon, tile_size) / tile_size)
        pt2_y = floor(lat2pixel(self.zoom, pt2.rlat, tile_size) / tile_size)
        for ptx_i in range(pt1_x, pt2_x):
            for pty_j in range(pt1_y, pt2_y):
                tile_path = "%s%c%d%c%d%c%d.%s" % (
                    map_path,
                    os.sep,
                    self.zoom,
                    os.sep,
                    ptx_i,
                    os.sep,
                    pty_j,
                    img_format,
                )
                _LOG.debug("file removed : %s", tile_path)
                try:
                    os.unlink(tile_path)
                except:
                    # The tile doesn't exist because it is in the load queue.
                    # That occurs when zooming.
                    pass
        self.osm.download_maps(pt1, pt2, self.zoom, self.zoom)

    def update_shortcuts(self, arg):
        """
        connect the keyboard or the keypad for shortcuts
        arg is mandatory because this function is also called by
        the checkbox button
        """
        dummy_arg = arg
        config.set("geography.use-keypad", self._config.get("geography.use-keypad"))
        if config.get("geography.use-keypad"):
            self.osm.set_keyboard_shortcut(
                osmgpsmap.MapKey_t.ZOOMIN, Gdk.keyval_from_name("KP_Add")
            )
            self.osm.set_keyboard_shortcut(
                osmgpsmap.MapKey_t.ZOOMOUT, Gdk.keyval_from_name("KP_Subtract")
            )
        else:
            self.osm.set_keyboard_shortcut(
                osmgpsmap.MapKey_t.ZOOMIN, Gdk.keyval_from_name("plus")
            )
            self.osm.set_keyboard_shortcut(
                osmgpsmap.MapKey_t.ZOOMOUT, Gdk.keyval_from_name("minus")
            )

        self.osm.set_keyboard_shortcut(
            osmgpsmap.MapKey_t.UP, Gdk.keyval_from_name("Up")
        )
        self.osm.set_keyboard_shortcut(
            osmgpsmap.MapKey_t.DOWN, Gdk.keyval_from_name("Down")
        )
        self.osm.set_keyboard_shortcut(
            osmgpsmap.MapKey_t.LEFT, Gdk.keyval_from_name("Left")
        )
        self.osm.set_keyboard_shortcut(
            osmgpsmap.MapKey_t.RIGHT, Gdk.keyval_from_name("Right")
        )

        # For shortcuts work, we must grab the focus
        self.osm.grab_focus()

    def add_selection_layer(self):
        """
        add the selection layer
        """
        selection_layer = SelectionLayer()
        self.osm.layer_add(selection_layer)
        return selection_layer

    def get_selection_layer(self):
        """
        get the selection layer
        """
        return self.selection_layer

    def add_kml_layer(self):
        """
        add the kml layer to the map
        """
        kml_layer = KmlLayer()
        self.osm.layer_add(kml_layer)
        return kml_layer

    def get_kml_layer(self):
        """
        get the kml layer
        """
        return self.kml_layer

    def add_message_layer(self):
        """
        add the message layer to the map
        """
        message_layer = MessageLayer()
        self.osm.layer_add(message_layer)
        return message_layer

    def get_message_layer(self):
        """
        get the message layer
        """
        return self.message_layer

    def add_date_layer(self):
        """
        add the date layer to the map
        """
        date_layer = DateLayer()
        self.osm.layer_add(date_layer)
        return date_layer

    def get_date_layer(self):
        """
        get the date layer
        """
        return self.date_layer

    def add_marker_layer(self):
        """
        add the marker layer
        """
        marker_layer = MarkerLayer()
        self.osm.layer_add(marker_layer)
        return marker_layer

    def get_marker_layer(self):
        """
        get the marker layer
        """
        return self.marker_layer

    def add_lifeway_layer(self):
        """
        add the track or life ways layer
        """
        lifeway_layer = LifeWayLayer()
        self.osm.layer_add(lifeway_layer)
        return lifeway_layer

    def get_lifeway_layer(self):
        """
        get the track or life ways layer
        """
        return self.lifeway_layer

    def remove_layer(self, layer):
        """
        remove the specified layer
        """
        self.osm.layer_remove(layer)

    def zoom_changed(self, zoom):
        """
        save the zoom and the position
        """
        dummy_zoom = zoom
        config.set("geography.zoom", self.osm.props.zoom)
        self.save_center(self.osm.props.latitude, self.osm.props.longitude)

    def motion_event(self, osmmap, event):
        """
        Moving during selection
        """
        current = osmmap.convert_screen_to_geographic(int(event.x), int(event.y))
        lat, lon = current.get_degrees()
        if self.zone_selection:
            # We draw a rectangle to show the selected region.
            layer = self.get_selection_layer()
            if layer:
                self.osm.layer_remove(layer)
            self.selection_layer = self.add_selection_layer()
            if self.end_selection is None:
                self.selection_layer.add_rectangle(self.begin_selection, current)
            else:
                self.selection_layer.add_rectangle(
                    self.begin_selection, self.end_selection
                )
        else:
            places = self.is_there_a_place_here(lat, lon)
            mess = ""
            for plc in places:
                if mess != "":
                    mess += " || "
                mess += plc[0]
            self.uistate.status.pop(self.context_id)
            self.context_id = self.uistate.status.push(1, mess)

    def save_center(self, lat, lon):
        """
        Save the longitude and lontitude in case we switch between maps.
        """
        _LOG.debug("save_center : %s,%s", lat, lon)
        if (-90.0 < lat < +90.0) and (-180.0 < lon < +180.0):
            config.set("geography.center-lat", lat)
            config.set("geography.center-lon", lon)
        else:
            _LOG.debug("save_center : new coordinates : %s,%s", lat, lon)
            _LOG.debug("save_center : old coordinates : %s,%s", lat, lon)
            # osmgpsmap bug ? reset to prior values to avoid osmgpsmap problems.
            self.osm.set_center_and_zoom(
                config.get("geography.center-lat"),
                config.get("geography.center-lon"),
                config.get("geography.zoom"),
            )

    def activate_selection_zoom(self, osm, event):
        """
        Zoom when in zone selection
        """
        dummy_osm = osm
        dummy_evt = event
        if self.end_selection is not None:
            self._autozoom()
        return True

    def map_clicked(self, osm, event):
        """
        Someone click on the map. Look at if we have a marker.
        mouse button 1 : zone selection or marker selection
        mouse button 2 : begin zone selection
        mouse button 3 : call the menu
        """
        self.osm.grab_focus()
        lat, lon = self.osm.get_event_location(event).get_degrees()
        current = osm.convert_screen_to_geographic(int(event.x), int(event.y))
        lat, lon = current.get_degrees()
        if event.button == 1 and event.type == Gdk.EventType.BUTTON_PRESS:
            if self.end_selection is not None:
                self.activate_selection_zoom(osm, event)
                self.end_selection = None
            else:
                # do we click on a marker ?
                self.is_there_a_marker_here(event, lat, lon)
        elif event.button == 2 and event.type == Gdk.EventType.BUTTON_PRESS:
            self.begin_selection = current
            self.end_selection = None
            self.zone_selection = True
        elif event.button == 2 and event.type == Gdk.EventType.BUTTON_RELEASE:
            self.end_selection = current
            self.zone_selection = False
        elif Gdk.Event.triggers_context_menu(event):
            self.build_nav_menu(osm, event, lat, lon)
        else:
            self.save_center(lat, lon)

    def is_there_a_place_here(self, lat, lon):
        """
        Is there a place at this position ?
        If too many places, this function is very time consuming
        """
        mark_selected = []
        if self.no_show_places_in_status_bar:
            return mark_selected
        oldplace = ""
        _LOG.debug(
            "%s",
            time.strftime(
                "start is_there_a_place_here : " "%a %d %b %Y %H:%M:%S", time.gmtime()
            ),
        )
        for mark in self.places_found:
            # as we are not precise with our hand, reduce the precision
            # depending on the zoom.
            if mark[0] != oldplace:
                oldplace = mark[0]
                precision = {
                    1: "%3.0f",
                    2: "%3.1f",
                    3: "%3.1f",
                    4: "%3.1f",
                    5: "%3.2f",
                    6: "%3.2f",
                    7: "%3.2f",
                    8: "%3.3f",
                    9: "%3.3f",
                    10: "%3.3f",
                    11: "%3.3f",
                    12: "%3.3f",
                    13: "%3.3f",
                    14: "%3.4f",
                    15: "%3.4f",
                    16: "%3.4f",
                    17: "%3.4f",
                    18: "%3.4f",
                }.get(config.get("geography.zoom"), "%3.1f")
                shift = {
                    1: 5.0,
                    2: 5.0,
                    3: 3.0,
                    4: 1.0,
                    5: 0.5,
                    6: 0.3,
                    7: 0.15,
                    8: 0.06,
                    9: 0.03,
                    10: 0.015,
                    11: 0.005,
                    12: 0.003,
                    13: 0.001,
                    14: 0.0005,
                    15: 0.0003,
                    16: 0.0001,
                    17: 0.0001,
                    18: 0.0001,
                }.get(config.get("geography.zoom"), 5.0)
                latp = precision % lat
                lonp = precision % lon
                mlatp = precision % float(mark[1])
                mlonp = precision % float(mark[2])
                latok = lonok = False
                if (float(mlatp) >= (float(latp) - shift)) and (
                    float(mlatp) <= (float(latp) + shift)
                ):
                    latok = True
                if (float(mlonp) >= (float(lonp) - shift)) and (
                    float(mlonp) <= (float(lonp) + shift)
                ):
                    lonok = True
                if latok and lonok:
                    mark_selected.append(mark)
        _LOG.debug(
            "%s",
            time.strftime(
                "  end is_there_a_place_here : " "%a %d %b %Y %H:%M:%S", time.gmtime()
            ),
        )
        return mark_selected

    def build_nav_menu(self, osm, event, lat, lon):
        """
        Must be implemented in the caller class
        """
        raise NotImplementedError

    def is_there_a_marker_here(self, event, lat, lon):
        """
        Must be implemented in the caller class
        """
        raise NotImplementedError

    def set_crosshair(self, active):
        """
        Show or hide the crosshair ?
        """
        if active:
            self.osm.layer_remove(self.cross_map)
            self.cross_map = osmgpsmap.MapOsd(show_crosshair=True)
            self.osm.layer_add(self.cross_map)
        else:
            self.osm.layer_remove(self.cross_map)
            self.cross_map = osmgpsmap.MapOsd(show_crosshair=False)
            self.osm.layer_add(self.cross_map)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2022       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
A ciation formatter that implements the current functionality.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.lib import Citation
from gramps.gen.utils.string import conf_strings

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
LOG = logging.getLogger(".cite")


class Formatter:
    def __init__(self):
        pass

    def format(self, bibliography, db, elocale):
        result = []
        for citation in bibliography.get_citation_list():
            source = db.get_source_from_handle(citation.get_source_handle())
            src = _format_source_text(source, elocale)

            refs = []
            for key, ref in citation.get_ref_list():
                text = _format_ref_text(ref, key, elocale)
                refs.append([ref, key, text])

            result.append([[source, src], refs])

        return result


def _format_source_text(source, elocale):
    if not source:
        return ""

    trans_text = elocale.translation.gettext
    # trans_text is a defined keyword (see po/update_po.py, po/genpot.sh)

    src_txt = ""

    if source.get_author():
        src_txt += source.get_author()

    if source.get_title():
        if src_txt:
            # Translators: needed for Arabic, ignore otherwise
            src_txt += trans_text(", ")
        # Translators: used in French+Russian, ignore otherwise
        src_txt += trans_text('"%s"') % source.get_title()

    if source.get_publication_info():
        if src_txt:
            # Translators: needed for Arabic, ignore otherwise
            src_txt += trans_text(", ")
        src_txt += source.get_publication_info()

    if source.get_abbreviation():
        if src_txt:
            # Translators: needed for Arabic, ignore otherwise
            src_txt += trans_text(", ")
        src_txt += "(%s)" % source.get_abbreviation()

    return src_txt


def _format_ref_text(ref, key, elocale):
    if not ref:
        return ""

    ref_txt = ""

    datepresent = False
    date = ref.get_date_object()
    if date is not None and not date.is_empty():
        datepresent = True
    if datepresent:
        if ref.get_page():
            ref_txt = "%s - %s" % (ref.get_page(), elocale.get_date(date))
        else:
            ref_txt = elocale.get_date(date)
    else:
        ref_txt = ref.get_page()

    # Print only confidence level if it is not Normal
    if ref.get_confidence_level() != Citation.CONF_NORMAL:
        ref_txt += (
            " ["
            + elocale.translation.gettext(conf_strings[ref.get_confidence_level()])
            + "]"
        )

    return ref_txt



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2009   Stephane Charette
# Copyright (C) 2008        Brian Matherly
# Copyright (C) 2010        Jakim Friant
# Copyright (C) 2010        Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Find people who are not related to the selected person"

# ------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# ------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
ngettext = glocale.translation.ngettext  # else "nearby" comments are ignored
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gen.errors import WindowActiveError
from gramps.gui.plug import tool
from gramps.gen.plug.report import utils
from gramps.gui.editors import EditPerson, EditFamily
from gramps.gui.managedwindow import ManagedWindow
from gramps.gui.utils import ProgressMeter
from gramps.gui.display import display_help
from gramps.gui.glade import Glade
from gramps.gen.lib import Tag
from gramps.gen.db import DbTxn

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Tools" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Not_Related", "manual")


# ------------------------------------------------------------------------
#
# NotRelated class
#
# ------------------------------------------------------------------------
class NotRelated(tool.ActivePersonTool, ManagedWindow):
    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate
        tool.ActivePersonTool.__init__(self, dbstate, uistate, options_class, name)

        if self.fail:  # bug #2709 -- fail if we have no active person
            return

        person_handle = uistate.get_active("Person")
        person = dbstate.db.get_person_from_handle(person_handle)
        self.name = person.get_primary_name().get_regular_name()
        self.title = _('Not related to "%s"') % self.name
        ManagedWindow.__init__(self, uistate, [], self.__class__)
        self.dbstate = dbstate
        self.uistate = uistate
        self.db = dbstate.db

        topDialog = Glade()

        topDialog.connect_signals(
            {
                "destroy_passed_object": self.close,
                "on_help_clicked": self.on_help_clicked,
                "on_delete_event": self.close,
            }
        )

        window = topDialog.toplevel
        title = topDialog.get_object("title")
        self.set_window(window, title, self.title)
        self.setup_configs("interface.notrelated", 450, 400)

        self.tagcombo = topDialog.get_object("tagcombo")
        tagmodel = Gtk.ListStore(str)
        self.tagcombo.set_model(tagmodel)
        self.tagcombo.set_entry_text_column(0)
        tagmodel.append((_("ToDo"),))
        tagmodel.append((_("NotRelated"),))
        self.tagcombo.set_sensitive(False)

        self.tagapply = topDialog.get_object("tagapply")
        self.tagapply.set_sensitive(False)
        self.tagapply.connect("clicked", self.applyTagClicked)

        # start the progress indicator
        self.progress = ProgressMeter(
            self.title, _("Starting"), parent=self.uistate.window
        )

        # setup the columns
        self.model = Gtk.TreeStore(
            GObject.TYPE_STRING,  # 0==name
            GObject.TYPE_STRING,  # 1==person gid
            GObject.TYPE_STRING,  # 2==parents
            GObject.TYPE_STRING,  # 3==tags
            GObject.TYPE_STRING,
        )  # 4==family gid (not shown to user)

        # note -- don't assign the model to the tree until it has been populated,
        # otherwise the screen updates are terribly slow while names are appended
        self.treeView = topDialog.get_object("treeview")
        col1 = Gtk.TreeViewColumn(_("Name"), Gtk.CellRendererText(), text=0)
        col2 = Gtk.TreeViewColumn(_("ID"), Gtk.CellRendererText(), text=1)
        col3 = Gtk.TreeViewColumn(_("Parents"), Gtk.CellRendererText(), text=2)
        col4 = Gtk.TreeViewColumn(_("Tags"), Gtk.CellRendererText(), text=3)
        col1.set_resizable(True)
        col2.set_resizable(True)
        col3.set_resizable(True)
        col4.set_resizable(True)
        col1.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col2.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col3.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col4.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col1.set_sort_column_id(0)
        #        col2.set_sort_column_id(1)
        #        col3.set_sort_column_id(2)
        col4.set_sort_column_id(3)
        self.treeView.append_column(col1)
        self.treeView.append_column(col2)
        self.treeView.append_column(col3)
        self.treeView.append_column(col4)
        self.treeSelection = self.treeView.get_selection()
        self.treeSelection.set_mode(Gtk.SelectionMode.MULTIPLE)
        self.treeSelection.set_select_function(self.selectIsAllowed, None)
        self.treeSelection.connect("changed", self.rowSelectionChanged)
        self.treeView.connect("row-activated", self.rowActivated)

        # initialize a few variables we're going to need
        self.numberOfPeopleInDatabase = self.db.get_number_of_people()
        self.numberOfRelatedPeople = 0
        self.numberOfUnrelatedPeople = 0

        # create the sets used to track related and unrelated people
        self.handlesOfPeopleToBeProcessed = set()
        self.handlesOfPeopleAlreadyProcessed = set()
        self.handlesOfPeopleNotRelated = set()

        # build a set of all people related to the selected person
        self.handlesOfPeopleToBeProcessed.add(person.get_handle())
        self.findRelatedPeople()

        # now that we have our list of related people, find everyone
        # in the database who isn't on our list
        self.findUnrelatedPeople()

        # populate the treeview model with the names of unrelated people
        if self.numberOfUnrelatedPeople == 0:
            # feature request 2356: avoid genitive form
            title.set_text(_("Everyone in the database is related to %s") % self.name)
        else:
            self.populateModel()
            self.model.set_sort_column_id(0, Gtk.SortType.ASCENDING)
            self.treeView.set_model(self.model)
            #            self.treeView.set_row_separator_func(self.iterIsSeparator, None)
            self.treeView.expand_all()

        # done searching through the database, so close the progress bar
        self.progress.close()

        self.show()

    def iterIsSeparator(self, model, iter):
        # return True only if the row is to be treated as a separator
        if self.model.get_value(iter, 1) == "":  # does the row have a GID?
            return True
        return False

    def selectIsAllowed(self, selection, model, path, isSelected, userData):
        # return True/False depending on if the row being selected is a leaf node
        iter = self.model.get_iter(path)
        if self.model.get_value(iter, 1) == "":  # does the row have a GID?
            return False
        return True

    def rowSelectionChanged(self, selection):
        state = selection.count_selected_rows() > 0
        self.tagcombo.set_sensitive(state)
        self.tagapply.set_sensitive(state)

    def rowActivated(self, treeView, path, column):
        # first we need to check that the row corresponds to a person
        iter = self.model.get_iter(path)
        personGid = self.model.get_value(iter, 1)
        familyGid = self.model.get_value(iter, 4)

        if familyGid != "":  # do we have a family?
            # get the parent family for this person
            family = self.db.get_family_from_gramps_id(familyGid)
            if family:
                try:
                    EditFamily(self.dbstate, self.uistate, [], family)
                except WindowActiveError:
                    pass

        elif personGid != "":  # do we have a person?
            # get the person that corresponds to this GID
            person = self.db.get_person_from_gramps_id(personGid)
            if person:
                try:
                    EditPerson(self.dbstate, self.uistate, [], person)
                except WindowActiveError:
                    pass

    def on_help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""
        display_help(WIKI_HELP_PAGE, WIKI_HELP_SEC)

    def applyTagClicked(self, button):
        progress = None
        rows = self.treeSelection.count_selected_rows()
        tag_name = str(self.tagcombo.get_active_text())

        # start the db transaction
        with DbTxn("Tag not related", self.db) as transaction:
            tag = self.db.get_tag_from_name(tag_name)
            if not tag:
                # create the tag if it doesn't already exist
                tag = Tag()
                tag.set_name(tag_name)
                tag.set_priority(self.db.get_number_of_tags())
                tag_handle = self.db.add_tag(tag, transaction)
            else:
                tag_handle = tag.get_handle()

            # if more than 1 person is selected, use a progress indicator
            if rows > 1:
                progress = ProgressMeter(self.title, _("Starting"), parent=self.window)
                progress.set_pass(
                    # Translators: leave all/any {...} untranslated
                    # TRANS: no singular form needed, as rows is always > 1
                    ngettext(
                        "Setting tag for {number_of} person",
                        "Setting tag for {number_of} people",
                        rows,
                    ).format(number_of=rows),
                    rows,
                )

            # iterate through all of the selected rows
            (model, paths) = self.treeSelection.get_selected_rows()

            for path in paths:
                if progress:
                    progress.step()

                # for the current row, get the GID and the person from the database
                iter = self.model.get_iter(path)
                personGid = self.model.get_value(iter, 1)
                person = self.db.get_person_from_gramps_id(personGid)

                # add the tag to the person
                person.add_tag(tag_handle)

                # save this change
                self.db.commit_person(person, transaction)

        # refresh the tags column
        self.treeView.set_model(None)
        for path in paths:
            iter = self.model.get_iter(path)
            personGid = self.model.get_value(iter, 1)
            person = self.db.get_person_from_gramps_id(personGid)
            self.model.set_value(iter, 3, self.get_tag_list(person))
        self.treeView.set_model(self.model)
        self.treeView.expand_all()

        if progress:
            progress.close()

    def findRelatedPeople(self):
        self.progress.set_pass(
            # Translators: leave all/any {...} untranslated
            # TRANS: No singular form is needed.
            ngettext(
                "Finding relationships between {number_of} person",
                "Finding relationships between {number_of} people",
                self.numberOfPeopleInDatabase,
            ).format(number_of=self.numberOfPeopleInDatabase),
            self.numberOfPeopleInDatabase,
        )

        # as long as we have people we haven't processed yet, keep looping
        while len(self.handlesOfPeopleToBeProcessed) > 0:
            handle = self.handlesOfPeopleToBeProcessed.pop()

            ### DEBUG DEBUG DEBUG
            #            if len(self.handlesOfPeopleAlreadyProcessed) > 50:
            #                break
            ###

            # see if we've already processed this person
            if handle in self.handlesOfPeopleAlreadyProcessed:
                continue

            person = self.db.get_person_from_handle(handle)

            # if we get here, then we're dealing with someone new
            self.progress.step()

            # remember that we've now seen this person
            self.handlesOfPeopleAlreadyProcessed.add(handle)

            # we have 4 things to do:  find (1) spouses, (2) parents, siblings(3), and (4) children

            # step 1 -- spouses
            for familyHandle in person.get_family_handle_list():
                family = self.db.get_family_from_handle(familyHandle)
                spouseHandle = utils.find_spouse(person, family)
                if (
                    spouseHandle
                    and spouseHandle not in self.handlesOfPeopleAlreadyProcessed
                ):
                    self.handlesOfPeopleToBeProcessed.add(spouseHandle)

            # step 2 -- parents
            for familyHandle in person.get_parent_family_handle_list():
                family = self.db.get_family_from_handle(familyHandle)
                fatherHandle = family.get_father_handle()
                motherHandle = family.get_mother_handle()
                if (
                    fatherHandle
                    and fatherHandle not in self.handlesOfPeopleAlreadyProcessed
                ):
                    self.handlesOfPeopleToBeProcessed.add(fatherHandle)
                if (
                    motherHandle
                    and motherHandle not in self.handlesOfPeopleAlreadyProcessed
                ):
                    self.handlesOfPeopleToBeProcessed.add(motherHandle)

            # step 3 -- siblings
            for familyHandle in person.get_parent_family_handle_list():
                family = self.db.get_family_from_handle(familyHandle)
                for childRef in family.get_child_ref_list():
                    childHandle = childRef.ref
                    if (
                        childHandle
                        and childHandle not in self.handlesOfPeopleAlreadyProcessed
                    ):
                        self.handlesOfPeopleToBeProcessed.add(childHandle)

            # step 4 -- children
            for familyHandle in person.get_family_handle_list():
                family = self.db.get_family_from_handle(familyHandle)
                for childRef in family.get_child_ref_list():
                    childHandle = childRef.ref
                    if (
                        childHandle
                        and childHandle not in self.handlesOfPeopleAlreadyProcessed
                    ):
                        self.handlesOfPeopleToBeProcessed.add(childHandle)

    def findUnrelatedPeople(self):
        # update our numbers
        self.numberOfRelatedPeople = len(self.handlesOfPeopleAlreadyProcessed)
        self.numberOfUnrelatedPeople = (
            self.numberOfPeopleInDatabase - self.numberOfRelatedPeople
        )

        if self.numberOfUnrelatedPeople > 0:
            # we have at least 1 "unrelated" person to find

            self.progress.set_pass(
                # Translators: leave all/any {...} untranslated
                ngettext(
                    "Looking for {number_of} person",
                    "Looking for {number_of} people",
                    self.numberOfUnrelatedPeople,
                ).format(number_of=self.numberOfUnrelatedPeople),
                self.numberOfPeopleInDatabase,
            )

            # loop through everyone in the database
            for handle in self.db.iter_person_handles():
                self.progress.step()

                # if this person is related, then skip to the next one
                if handle in self.handlesOfPeopleAlreadyProcessed:
                    continue

                ### DEBUG DEBUG DEBUG
                #                if len(self.handlesOfPeopleNotRelated) > 10:
                #                    break
                ###

                # if we get here, we have someone who is "not related"
                self.handlesOfPeopleNotRelated.add(handle)

    def populateModel(self):
        self.progress.set_pass(
            # Translators: leave all/any {...} untranslated
            ngettext(
                "Looking up the name of {number_of} person",
                "Looking up the names of {number_of} people",
                self.numberOfUnrelatedPeople,
            ).format(number_of=self.numberOfUnrelatedPeople),
            self.numberOfUnrelatedPeople,
        )

        # loop through the entire list of unrelated people
        for handle in self.handlesOfPeopleNotRelated:
            self.progress.step()
            person = self.db.get_person_from_handle(handle)
            primaryname = person.get_primary_name()
            surname = primaryname.get_surname()
            name = primaryname.get_name()
            gid = person.get_gramps_id()

            # Retrieve the sorted tag list
            tag_list = self.get_tag_list(person)

            # find the names of the parents
            familygid = ""
            parentNames = ""
            parentFamilyHandle = person.get_main_parents_family_handle()
            if parentFamilyHandle:
                parentFamily = self.db.get_family_from_handle(parentFamilyHandle)
                familygid = parentFamily.get_gramps_id()
                fatherName = None
                motherName = None
                fatherHandle = parentFamily.get_father_handle()
                if fatherHandle:
                    father = self.db.get_person_from_handle(fatherHandle)
                    fatherName = father.get_primary_name().get_first_name()
                motherHandle = parentFamily.get_mother_handle()
                if motherHandle:
                    mother = self.db.get_person_from_handle(motherHandle)
                    motherName = mother.get_primary_name().get_first_name()

                # now that we have the names, come up with a label we can use
                if fatherName:
                    parentNames += fatherName
                if fatherName and motherName:
                    parentNames += " & "
                if motherName:
                    parentNames += motherName

            # get the surname node (or create it if it doesn't exist)

            # start with the root
            iter = self.model.get_iter_first()
            # look for a node with a matching surname
            while iter:
                if self.model.get_value(iter, 0) == surname:
                    break
                iter = self.model.iter_next(iter)

            # if we don't have a valid iter, then create a new top-level node
            if not iter:
                iter = self.model.append(None, [surname, "", "", "", ""])

            # finally, we now get to add this person to the model
            self.model.append(iter, [name, gid, parentNames, tag_list, familygid])

    def build_menu_names(self, obj):
        return (self.title, None)

    def get_tag_list(self, person):
        """
        Return a sorted list of tag names for the given person.
        """
        tags = []
        for handle in person.get_tag_list():
            tag = self.db.get_tag_from_handle(handle)
            tags.append(tag.get_name())
        tags.sort(key=glocale.sort_key)
        # TODO for Arabic, should the next line's comma be translated?
        return ", ".join(tags)


# ------------------------------------------------------------------------
#
# NotRelatedOptions
#
# ------------------------------------------------------------------------
class NotRelatedOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        """Initialize the options class"""
        tool.ToolOptions.__init__(self, name, person_id)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2007  Donald N. Allingham, Martin Hawlisch
# Copyright (C) 2009 Douglas S. Blank
# Copyright (C) 2012 Benny Malengier
# Copyright (C) 2013 Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
# from gi.repository import Pango
# from gi.repository import GObject
# from gi.repository import Gdk
# from gi.repository import Gtk
# from gi.repository import PangoCairo
import cairo

# import math
# import colorsys
# import pickle


# -------------------------------------------------------------------------
#
# Functions
#
# -------------------------------------------------------------------------
def warpPath(ctx, function):
    """Transform a path given a 2D transformation function.

    ctx -- a cairo.Context, on which the path is set
    function -- a 2D transform (x,y) |-> (x_new,y_new)

    The transformed path replaces the original one on the context.

    Taken from /usr/share/doc/python-cairo/examples/warpedtext.py
    According to /usr/share/doc/python-cairo/copyright, licensed
    under MOZILLA PUBLIC LICENSE 1.1, see that file for more detail.
    """

    first = True

    for type, points in ctx.copy_path():
        if type == cairo.PATH_MOVE_TO:
            if first:
                ctx.new_path()
                first = False
            x, y = function(*points)
            ctx.move_to(x, y)

        elif type == cairo.PATH_LINE_TO:
            x, y = function(*points)
            ctx.line_to(x, y)

        elif type == cairo.PATH_CURVE_TO:
            x1, y1, x2, y2, x3, y3 = points
            x1, y1 = function(x1, y1)
            x2, y2 = function(x2, y2)
            x3, y3 = function(x3, y3)
            ctx.curve_to(x1, y1, x2, y2, x3, y3)

        elif type == cairo.PATH_CLOSE_PATH:
            ctx.close_path()



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Slovenian version 2010 by Bernard Banko, based on croatian one by Josip

"""
Slovenian-specific classes for parsing and displaying dates - new framework.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Slovenian parser
#
# -------------------------------------------------------------------------
class DateParserSL(DateParser):
    """
    Converts a text string into a Date object
    """

    modifier_to_int = {
        "pred": Date.MOD_BEFORE,
        "pr.": Date.MOD_BEFORE,
        "po": Date.MOD_AFTER,
        "okoli": Date.MOD_ABOUT,
        "okrog": Date.MOD_ABOUT,
        "okr.": Date.MOD_ABOUT,
        "ok.": Date.MOD_ABOUT,
        "cca.": Date.MOD_ABOUT,
        "cca": Date.MOD_ABOUT,
        "circa": Date.MOD_ABOUT,
        "ca.": Date.MOD_ABOUT,
        "približno": Date.MOD_ABOUT,
        "pribl.": Date.MOD_ABOUT,
        "~": Date.MOD_ABOUT,
        "od": Date.MOD_FROM,
        "do": Date.MOD_TO,
    }

    quality_to_int = {
        "ocenjeno": Date.QUAL_ESTIMATED,
        "oc.": Date.QUAL_ESTIMATED,
        "po oceni": Date.QUAL_ESTIMATED,
        "izračunano": Date.QUAL_CALCULATED,
        "izrač.": Date.QUAL_CALCULATED,
        "po izračunu": Date.QUAL_CALCULATED,
    }

    bce = [
        "pred našim štetjem",
        "pred Kristusom",
        "p.n.š.",
        "p. n. š.",
        "pr.Kr.",
        "pr. Kr.",
    ] + DateParser.bce

    def init_strings(self):
        """
        compiles regular expression strings for matching dates
        """

        DateParser.init_strings(self)
        # match 'Day. MONTH year.' format with or without dots
        self._text2 = re.compile(
            r"(\d+)?\.?\s*?%s\.?\s*((\d+)(/\d+)?)?\s*\.?$" % self._mon_str,
            re.IGNORECASE,
        )
        # match Day.Month.Year.
        self._numeric = re.compile(r"((\d+)[/\.-])?\s*((\d+)[/\.-])?\s*(\d+)\.?$")

        self._span = re.compile(
            r"od\s+(?P<start>.+)\s+do\s+(?P<stop>.+)", re.IGNORECASE
        )
        self._range = re.compile(
            r"med\s+(?P<start>.+)\s+in\s+(?P<stop>.+)", re.IGNORECASE
        )
        self._jtext2 = re.compile(
            r"(\d+)?.?\s+?%s\s*((\d+)(/\d+)?)?" % self._jmon_str, re.IGNORECASE
        )


# -------------------------------------------------------------------------
#
# Slovenian display
#
# -------------------------------------------------------------------------
class DateDisplaySL(DateDisplay):
    """
    Slovenian language date display class.
    """

    # TODO fix BUG 7064: non-Gregorian calendars wrongly use BCE notation for negative dates
    # not refactoring _bce_str into base class because it'll be gone under #7064
    _bce_str = "%s pr.Kr."

    display = DateDisplay.display_formatted


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("sl_SI", "sl", "SL", "slovenščina", "slovenian", "Slovenian", ("%d. %m. %Y",)),
    DateParserSL,
    DateDisplaySL,
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Marker types.

From version 3.3 onwards, this is only kept to convert markers into tags
when loading old database files.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .grampstype import GrampsType
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class MarkerType(GrampsType):
    """
    Class for handling data markers.
    """

    NONE = -1
    CUSTOM = 0
    COMPLETE = 1
    TODO_TYPE = 2

    _CUSTOM = CUSTOM
    _DEFAULT = NONE

    _DATAMAP = [
        (NONE, "", ""),
        (CUSTOM, _("Custom"), "Custom"),
        (COMPLETE, _("Complete"), "Complete"),
        (TODO_TYPE, _("ToDo"), "ToDo"),
    ]

    def __init__(self, value=None):
        GrampsType.__init__(self, value)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2007  Donald N. Allingham
# Copyright (C) 2017       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Class handling language-specific selection for date parser and displayer.
"""

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..utils.grampslocale import GrampsLocale
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
# import prerequisites for localized handlers
from ._datehandler import (
    LANG,
    LANG_SHORT,
    LANG_TO_PARSER,
    LANG_TO_DISPLAY,
    locale_tformat,
    main_locale,
)
from . import _datestrings

# Import all the localized handlers
from . import _date_ar
from . import _date_bg
from . import _date_ca
from . import _date_cs
from . import _date_da
from . import _date_de
from . import _date_el
from . import _date_es
from . import _date_fi
from . import _date_fr
from . import _date_he
from . import _date_hr
from . import _date_hu
from . import _date_is
from . import _date_it
from . import _date_ja
from . import _date_lt
from . import _date_nb
from . import _date_nl
from . import _date_pl
from . import _date_pt
from . import _date_ru
from . import _date_sk
from . import _date_sl
from . import _date_sr
from . import _date_sv
from . import _date_uk
from . import _date_zh_CN
from . import _date_zh_TW

# the following makes sure we use the LC_TIME value for date display & parsing
dlocale = GrampsLocale(lang=glocale.calendar)


# Initialize global parser
try:
    if LANG in LANG_TO_PARSER:
        parser = LANG_TO_PARSER[LANG](plocale=dlocale)
    else:
        parser = LANG_TO_PARSER[LANG_SHORT](plocale=dlocale)
except:
    logging.warning(_("Date parser for '%s' not available, using default") % LANG)
    parser = LANG_TO_PARSER["C"](plocale=dlocale)

# Initialize global displayer
try:
    from ..config import config

    val = config.get("preferences.date-format")
except:
    val = 0

try:
    if LANG in LANG_TO_DISPLAY:
        displayer = LANG_TO_DISPLAY[LANG](val, blocale=dlocale)
    else:
        displayer = LANG_TO_DISPLAY[LANG_SHORT](val, blocale=dlocale)
except:
    logging.warning(_("Date displayer for '%s' not available, using default") % LANG)
    displayer = LANG_TO_DISPLAY["C"](val, blocale=dlocale)


# Import utility functions
from ._dateutils import *

# set GRAMPS_RESOURCES then: python3 -m gramps.gen.datehandler.__init__
if __name__ == "__main__":
    from ._datedisplay import DateDisplay

    m = 0
    date_handlers = sorted(LANG_TO_DISPLAY.items())
    for l, d in date_handlers:
        if len(l) != 2 and l not in ("zh_TW"):  # Chinese has two date_handlers
            continue
        if l.upper() == l and (l.lower(), d) in date_handlers:
            continue  # don't need to see the upper-case variant also
        m = max(m, len(d.formats))
        print(
            "{}: {} {} own-f:{} own-dc:{} own-dg:{}".format(
                l,
                len(d.formats),
                d.formats,
                d.formats != DateDisplay.formats,
                d._display_calendar != DateDisplay._display_calendar,
                d._display_gregorian != DateDisplay._display_gregorian,
            )
        )
    print("MAX: ", m)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# gen/mime/_pythonmime.py

import mimetypes
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

_type_map = {
    "application/x-gramps": "Gramps database",
    "application/x-gramps-xml": "Gramps XML database",
    "application/x-gedcom": "GEDCOM database",
    "application/x-gramps-package": "Gramps package",
    "image/jpeg": "JPEG image",
    "image/tiff": "TIFF image",
    "image/png": "PNG image",
    "application/pdf": "PDF document",
    "text/rtf": "Rich Text File",
}

mimetypes.add_type("application/x-gramps", ".grdb")
mimetypes.add_type("application/x-gramps", ".GRDB")
mimetypes.add_type("application/x-gramps-xml", ".gramps")
mimetypes.add_type("application/x-gramps-xml", ".Gramps")
mimetypes.add_type("application/x-gedcom", ".ged")
mimetypes.add_type("application/x-gedcom", ".GED")
mimetypes.add_type("application/x-gramps-package", ".gpkg")
mimetypes.add_type("application/x-gramps-package", ".GPKG")
mimetypes.add_type("text/x-comma-separated-values", ".csv")


def get_description(mime_type):
    """Return the description of the specified mime type"""
    return _type_map.get(mime_type, _("unknown"))


def get_type(filename):
    """Return the mime type of the specified file"""
    value = mimetypes.guess_type(filename)
    if value and value[0]:
        return value[0]
    else:
        return _("unknown")


def mime_type_is_defined(mime_type):
    """
    Return True if a description for a mime type exists.
    """
    return mime_type in _type_map
